"""Tests for synthetic probability loss functions."""

#########################################################################################################
# Imports

import unittest

import numpy as np
import tensorflow as tf

from tensiometer.synthetic_probability import loss_functions as lf

#########################################################################################################
# Test cases


class TestLossFunctions(unittest.TestCase):
    """Loss functions test suite."""
    def setUp(self):
        """Set up test fixtures."""
        self.y_true = tf.constant([[0.5], [1.0]], dtype=tf.float32)
        self.y_pred = tf.constant([[0.4], [0.8]], dtype=tf.float32)
        self.sample_weight = tf.constant([1.0, 2.0], dtype=tf.float32)

    def test_standard_loss(self):
        """Test standard loss."""
        loss = lf.standard_loss()
        val = loss(self.y_true, self.y_pred, sample_weight=self.sample_weight)
        self.assertTrue(np.all(np.isfinite(val.numpy())))
        components = loss.compute_loss_components(self.y_true, self.y_pred, self.sample_weight)
        self.assertTrue(np.all(np.isfinite(components.numpy())))
        loss.print_feedback(padding='')
        loss.reset()

    def test_constant_weight_loss(self):
        """Test constant weight loss."""
        loss = lf.constant_weight_loss(alpha=0.7, beta=0.1)
        val = loss(self.y_true, self.y_pred, sample_weight=self.sample_weight)
        self.assertTrue(np.all(np.isfinite(val.numpy())))
        comp = loss.compute_loss_components(self.y_true, self.y_pred, self.sample_weight)
        self.assertEqual(len(comp), 2)
        loss.print_feedback(padding='')
        loss.reset()

    def test_constant_weight_loss_default_weights(self):
        """Test constant weight loss with default weights."""
        loss = lf.constant_weight_loss(alpha=0.7, beta=0.1)
        val = loss(self.y_true, self.y_pred)
        self.assertTrue(np.all(np.isfinite(val.numpy())))

    def test_constant_weight_loss_graph_branch(self):
        """Test constant weight loss graph branch."""
        loss = lf.constant_weight_loss(alpha=0.5, beta=0.0)
        @tf.function
        def _compute():
            """Wrap the loss call in a graph context."""
            return loss(self.y_true, self.y_pred, sample_weight=self.sample_weight)
        val = _compute()
        self.assertTrue(np.all(np.isfinite(val.numpy())))

    def test_variable_weight_loss(self):
        """Test variable weight loss."""
        loss = lf.variable_weight_loss(lambda_1=0.5, lambda_2=0.5, beta=0.0)
        val = loss(self.y_true, self.y_pred, sample_weight=self.sample_weight)
        self.assertTrue(np.all(np.isfinite(val.numpy())))
        comp = loss.compute_loss_components(self.y_true, self.y_pred, self.sample_weight, lambda_1=0.5, lambda_2=0.5)
        self.assertEqual(len(comp), 4)
        with self.assertRaises(NotImplementedError):
            loss.update_lambda_values_on_epoch_begin(epoch=1)
        with self.assertRaises(NotImplementedError):
            loss.print_feedback(padding='')
        loss.reset()

    def test_variable_weight_loss_default_weights(self):
        """Test variable weight loss with default weights."""
        loss = lf.variable_weight_loss(lambda_1=0.5, lambda_2=0.5, beta=0.0)
        val = loss(self.y_true, self.y_pred)
        self.assertTrue(np.all(np.isfinite(val.numpy())))

    def test_variable_weight_loss_defaults_and_graph(self):
        """Test variable weight loss defaults and graph branch."""
        loss = lf.variable_weight_loss(lambda_1=0.25, lambda_2=0.75, beta=0.0)
        comp = loss.compute_loss_components(self.y_true, self.y_pred, self.sample_weight)
        self.assertEqual(len(comp), 4)
        @tf.function
        def _compute():
            """Wrap the loss call in a graph context."""
            return loss(self.y_true, self.y_pred, sample_weight=self.sample_weight)
        val = _compute()
        self.assertTrue(np.all(np.isfinite(val.numpy())))

    def test_random_weight_loss(self):
        """Test random weight loss."""
        loss = lf.random_weight_loss(initial_random_epoch=0, lambda_1=0.5, beta=0.0)
        loss.update_lambda_values_on_epoch_begin(epoch=0)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertIn(lam1, [0.0, 1.0, 0.5])
        loss.update_lambda_values_on_epoch_begin(epoch=2)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertIn(lam1, [0.0, 1.0])
        loss.print_feedback(padding='')

    def test_annealed_weight_loss(self):
        """Test annealed weight loss."""
        loss = lf.annealed_weight_loss(anneal_epoch=0, lambda_1=1.0, beta=0.0, roll_off_nepoch=1)
        loss.update_lambda_values_on_epoch_begin(epoch=1)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertLess(lam1, 1.0)
        loss.print_feedback(padding='')

    def test_annealed_weight_loss_noop(self):
        """Test annealed weight loss no-op branch."""
        loss = lf.annealed_weight_loss(anneal_epoch=5, lambda_1=0.8, beta=0.0, roll_off_nepoch=1)
        loss.update_lambda_values_on_epoch_begin(epoch=2)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertAlmostEqual(lam1, 0.8)

    def test_softadapt_loss(self):
        """Test SoftAdapt loss."""
        loss = lf.SoftAdapt_weight_loss(tau=1.0, smoothing=True, smoothing_tau=2)
        logs = {'val_rho_loss': [1.0, 0.9], 'val_ee_loss': [1.0, 1.1]}
        loss.update_lambda_values_on_epoch_begin(epoch=1, logs=logs)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertGreaterEqual(lam1, 0.0)
        loss.print_feedback(padding='')

    def test_softadapt_zero_rate_and_no_smoothing(self):
        """Test SoftAdapt zero rate with no smoothing."""
        loss = lf.SoftAdapt_weight_loss(tau=1.0, smoothing=False)
        logs = {'val_rho_loss': [1.0], 'val_ee_loss': [1.0]}
        loss.update_lambda_values_on_epoch_begin(epoch=0, logs=logs)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertEqual(lam1, 1.0)
        loss.print_feedback(padding='')

    def test_softadapt_no_smoothing_nonzero_rate(self):
        """Test SoftAdapt no smoothing with nonzero rate."""
        loss = lf.SoftAdapt_weight_loss(tau=1.0, smoothing=False)
        logs = {'val_rho_loss': [1.0, 0.5], 'val_ee_loss': [1.0, 1.5]}
        loss.update_lambda_values_on_epoch_begin(epoch=2, logs=logs)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertGreaterEqual(lam1, 0.0)
        loss.print_feedback(padding='')

    def test_sharpstep_loss(self):
        """Test SharpStep loss."""
        loss = lf.SharpStep(step_epoch=1, value_1=1.0, value_2=0.25, beta=0.0)
        loss.update_lambda_values_on_epoch_begin(epoch=0)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertEqual(lam1, 1.0)
        loss.update_lambda_values_on_epoch_begin(epoch=2)
        lam1 = tf.keras.backend.get_value(loss.lambda_1)
        self.assertEqual(lam1, 0.25)
        loss.print_feedback(padding='')

    def test_broadcast_sample_weight_helpers(self):
        """Test broadcast sample weight helper branches."""
        losses = tf.constant([1.0, 2.0], dtype=tf.float32)
        self.assertIsNone(lf._broadcast_sample_weight(None, losses))
        scalar_weight = tf.constant(2.0, dtype=tf.float32)
        out = lf._broadcast_sample_weight(scalar_weight, losses)
        self.assertTrue(np.allclose(out.numpy(), np.array([2.0, 2.0], dtype=np.float32)))

        @tf.function(input_signature=[
            tf.TensorSpec(shape=None, dtype=tf.float32),
            tf.TensorSpec(shape=None, dtype=tf.float32),
        ])
        def _wrap(weight, tensor):
            """Wrap broadcast helper in graph mode."""
            return lf._broadcast_sample_weight(weight, tensor)

        weight = tf.constant([1.0, 2.0], dtype=tf.float32)
        loss_grid = tf.constant([[1.0, 2.0], [3.0, 4.0]], dtype=tf.float32)
        out = _wrap(weight, loss_grid)
        self.assertEqual(out.shape, loss_grid.shape)

    def test_broadcast_sample_weight_unknown_rank(self):
        """Test broadcast helper with unknown ranks."""
        losses = tf.constant([[1.0, 2.0], [3.0, 4.0]], dtype=tf.float32)
        weights = tf.constant([1.0, 2.0], dtype=tf.float32)
        losses.set_shape(tf.TensorShape(None))
        weights.set_shape(tf.TensorShape(None))
        out = lf._broadcast_sample_weight(weights, losses)
        expected = np.array([[1.0, 1.0], [2.0, 2.0]], dtype=np.float32)
        self.assertTrue(np.allclose(out.numpy(), expected))

    def test_broadcast_sample_weight_unknown_rank_tracing(self):
        """Test broadcast helper tracing with unknown ranks."""
        @tf.function(input_signature=[
            tf.TensorSpec(shape=None, dtype=tf.float32),
            tf.TensorSpec(shape=None, dtype=tf.float32),
        ])
        def _wrap(weight, tensor):
            """Wrap broadcast helper in tracing."""
            return lf._broadcast_sample_weight(weight, tensor)

        concrete = _wrap.get_concrete_function()
        out = concrete(
            tf.constant([1.0, 2.0], dtype=tf.float32),
            tf.constant([[1.0, 2.0], [3.0, 4.0]], dtype=tf.float32),
        )
        self.assertEqual(out.shape, (2, 2))

    def test_reduce_weighted_loss_helpers(self):
        """Test weighted loss reductions."""
        losses = tf.constant([1.0, 2.0], dtype=tf.float32)
        out_none = lf._reduce_weighted_loss(losses, None, tf.keras.losses.Reduction.NONE)
        self.assertTrue(np.allclose(out_none.numpy(), losses.numpy()))
        out_sum = lf._reduce_weighted_loss(losses, None, tf.keras.losses.Reduction.SUM)
        self.assertAlmostEqual(float(out_sum.numpy()), 3.0)
        out_mean = lf._reduce_weighted_loss(losses, None, tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE)
        self.assertAlmostEqual(float(out_mean.numpy()), 1.5)

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""
This file implements information geometry methods working on normalizing flows.

This file needs work and is not complete.
"""

###############################################################################
# initial imports and set-up:

import numpy as np
import scipy

# tensorflow imports:
import tensorflow as tf

###############################################################################
# CPC decomposition in tensorflow:


@tf.function
def tf_CPC_decomposition(matrix_a, matrix_b):
    """
    Covariant Principal Components decomposition impolemented in tensorflow.

    :param matrix_a: input matrix ``A``.
    :param matrix_b: input matrix ``B``.
    :returns: eigenvalues and eigenvectors of the transformed matrix.
    """
    # compute the eigenvalues of b, lambda_b:
    _lambda_b, _phi_b = tf.linalg.eigh(matrix_b)
    _sqrt_lambda_b = tf.linalg.diag(1./tf.math.sqrt(_lambda_b))
    _phib_prime = tf.matmul(_phi_b, _sqrt_lambda_b)
    #
    rank = tf.rank(_phib_prime)
    new_order = tf.concat([tf.range(rank - 2), [rank - 1, rank - 2]], axis=0)
    _phib_prime_T = tf.transpose(_phib_prime, perm=new_order)
    #
    _a_prime = tf.matmul(tf.matmul(_phib_prime_T, matrix_a), _phib_prime)
    _lambda, _phi_a = tf.linalg.eigh(_a_prime)
    _phi = tf.matmul(tf.matmul(_phi_b, _sqrt_lambda_b), _phi_a)
    #
    return _lambda, _phi

###############################################################################
# TensorFlow KL decomposition helper:


def tf_KL_decomposition(matrix_a, matrix_b):
    """
    TensorFlow implementation of the KL decomposition used within flow-based
    utilities. Mirrors :func:`tensiometer.utilities.stats_utilities.KL_decomposition`.

    :param matrix_a: first matrix ``A``.
    :param matrix_b: second matrix ``B`` (assumed positive definite).
    :returns: eigenvalues and eigenvectors of the generalized problem.
    """
    # compute the eigenvalues of b, lambda_b:
    _lambda_b, _phi_b = tf.linalg.eigh(matrix_b)
    _sqrt_lambda_b = tf.linalg.diag(1./tf.math.sqrt(_lambda_b))
    _phib_prime = tf.matmul(_phi_b, _sqrt_lambda_b)
    rank = tf.rank(_phib_prime)
    new_order = tf.concat([tf.range(rank - 2), [rank - 1, rank - 2]], axis=0)
    _phib_prime_T = tf.transpose(_phib_prime, perm=new_order)
    _a_prime = tf.matmul(tf.matmul(_phib_prime_T, matrix_a), _phib_prime)
    _lambda, _phi_a = tf.linalg.eigh(_a_prime)
    _phi = tf.matmul(tf.matmul(_phi_b, _sqrt_lambda_b), _phi_a)
    #
    return _lambda, _phi

###############################################################################
# Simple ODE solver:

# need here a simple ODE method

###############################################################################
# KL methods:


def _naive_eigenvalue_ode_abs(flow, t, y, reference):
    """
    Solve naively the dynamical equation for eigenvalues in abstract space.

    :param flow: flow object providing Jacobians.
    :param t: integration time.
    :param y: current position in abstract space.
    :param reference: reference eigenvector used to select the branch.
    :returns: eigenvector direction at the current step.
    """
    # preprocess:
    x = flow.cast([y])
    # map to original space to compute Jacobian (without inversion):
    x_par = flow.map_to_original_coord(x)
    # precompute Jacobian and its derivative:
    jac = flow.inverse_jacobian(x_par)[0]
    jac_T = tf.transpose(jac)
    jac_jac_T = tf.matmul(jac, jac_T)
    # compute eigenvalues:
    eig, eigv = tf.linalg.eigh(jac_jac_T)
    temp = tf.matmul(tf.transpose(eigv), tf.transpose([reference]))
    idx = tf.math.argmax(tf.abs(temp))[0]
    w = tf.convert_to_tensor([tf.math.sign(temp[idx]) * eigv[:, idx]])
    #
    return w


def solve_eigenvalue_ode_abs(self, y0, n, length=1.5, side='both', integrator_options=None, num_points=100, **kwargs):
    """
    Solve eigenvalue ODE in abstract space.

    :param y0: starting point in abstract coordinates.
    :param n: index of the eigenvector to follow.
    :param length: integration length for each direction.
    :param side: which direction to integrate; ``'+'``, ``'-'`` or ``'both'``.
    :param integrator_options: optional options forwarded to ``scipy.integrate.ode``.
    :param num_points: number of output points per direction.
    :returns: solution times, trajectory, and velocity along the path.
    """
    y0_np = np.asarray(y0, dtype=np.float64)
    # define solution points:
    solution_times = tf.linspace(0., length, num_points)
    # compute initial PCA:
    x_abs = self.cast([y0_np])
    x_par = self.map_to_original_coord(x_abs)
    jac = self.inverse_jacobian(x_par)[0]
    jac_T = tf.transpose(jac)
    jac_jac_T = tf.matmul(jac, jac_T)
    # compute eigenvalues at initial point:
    eig, eigv = tf.linalg.eigh(jac_jac_T)
    # solve forward:
    if side == '+' or side == 'both':
        # initialize solution:
        temp_sol_1 = np.zeros((num_points-1, self.num_params))
        temp_sol_dot_1 = np.zeros((num_points-1, self.num_params))
        # initialize forward integration:
        solver = scipy.integrate.ode(self._naive_eigenvalue_ode_abs)
        if integrator_options is not None:
            solver.set_integrator(**integrator_options)
        solver.set_initial_value(y0_np, 0.)
        reference = eigv[:, n]
        yt = y0_np.copy()
        yprime = reference
        # do the time steps:
        for ind, t in enumerate(solution_times[1:]):
            # set the reference:
            solver.set_f_params(reference)
            # advance solver:
            try:
                yt = solver.integrate(t)
                yprime = self._naive_eigenvalue_ode_abs(t, yt, reference)
            except:
                pass
            # update reference:
            reference = yprime[0]
            # save out:
            temp_sol_1[ind] = yt.copy()
            temp_sol_dot_1[ind] = yprime.numpy().copy()
        # return if needed:
        if side == '+':
            traj = np.concatenate((x_abs.numpy(), temp_sol_1))
            vel = np.concatenate(([eigv[:, n].numpy()], temp_sol_dot_1))
            return solution_times, traj, vel
    # solve backward:
    if side == '-' or side == 'both':
        # initialize solution:
        temp_sol_2 = np.zeros((num_points-1, self.num_params))
        temp_sol_dot_2 = np.zeros((num_points-1, self.num_params))
        # initialize backward integration:
        solver = scipy.integrate.ode(self._naive_eigenvalue_ode_abs)
        if integrator_options is not None:
            solver.set_integrator(**integrator_options)
        solver.set_initial_value(y0_np, 0.)
        reference = - eigv[:, n]
        yt = y0_np.copy()
        yprime = reference
        for ind, t in enumerate(solution_times[1:]):
            # set the reference:
            solver.set_f_params(reference)
            # advance solver:
            try:
                yt = solver.integrate(t)
                yprime = self._naive_eigenvalue_ode_abs(t, yt, reference)
            except:
                pass
            # update reference:
            reference = yprime[0]
            # save out:
            temp_sol_2[ind] = yt.copy()
            temp_sol_dot_2[ind] = yprime.numpy().copy()
        # return if needed:
        if side == '-':
            traj = np.concatenate((temp_sol_2[::-1], x_abs.numpy()))
            vel = np.concatenate((-temp_sol_dot_2[::-1], [eigv[:, n].numpy()]))
            return -solution_times, traj, vel
    # patch solutions:
    times = np.concatenate((-solution_times[::-1], solution_times[1:]))
    traj = np.concatenate((temp_sol_2[::-1], x_abs.numpy(), temp_sol_1))
    vel = np.concatenate((-temp_sol_dot_2[::-1], [eigv[:, n].numpy()], temp_sol_dot_1))
    #
    return times, traj, vel

def solve_eigenvalue_ode_par(self, y0, n, **kwargs):
    """
    Solve the eigenvalue ODE in parameter space.

    :param y0: starting point in parameter coordinates.
    :param n: index of the eigenvector to follow.
    :returns: times and mapped trajectory in parameter space.
    """
    # go to abstract space:
    x_abs = self.map_to_abstract_coord(self.cast([y0]))[0]
    # call solver:
    times, traj, vel = self.solve_eigenvalue_ode_abs(x_abs, n, **kwargs)
    # convert back:
    traj = self.map_to_original_coord(self.cast(traj))
    #
    return times, traj

# solve full transport in abstract space:
@tf.function()
def eigenvalue_ode_abs_temp_3(self, t, y):
    # unpack y:
    x = self.cast([y[:self.num_params]])
    w = self.cast([y[self.num_params:-1]])
    alpha = self.cast([y[-1]])
    # map to original space to compute Jacobian (without inversion):
    x_par = self.map_to_original_coord(x)
    # precompute Jacobian and its derivative:
    jac = self.inverse_jacobian(x_par)[0]
    # derivative of the Jacobian is not implemented; use zeros as a placeholder
    djac = tf.zeros((self.num_params, self.num_params, self.num_params),
                    dtype=jac.dtype)
    jacm1 = self.direct_jacobian(x_par)[0]
    jac_T = tf.transpose(jac)
    jac_jac_T = tf.matmul(jac, jac_T)
    Id = tf.eye(self.num_params, dtype=jac_jac_T.dtype)
    # select the eigenvector that we want to follow based on the solution to the continuity equation:
    eig, eigv = tf.linalg.eigh(jac_jac_T)
    idx = tf.math.argmax(tf.abs(tf.matmul(tf.transpose(eigv), tf.transpose(w))))[0]
    tilde_w = tf.convert_to_tensor([eigv[:, idx]])
    dot_J = tf.einsum('k, lk, ijl -> ji', tilde_w[0], jacm1, djac)
    # equation for alpha:
    alpha_dot = 2.*tf.matmul(tf.matmul(tilde_w, jac), tf.matmul(dot_J, tf.transpose(tilde_w)))
    # equation for wdot:
    wdot_lhs = (jac_jac_T - tf.matmul(tf.matmul(tilde_w, jac_jac_T), tf.transpose(tilde_w))*Id)
    wdot_rhs = tf.matmul(alpha_dot - tf.matmul(dot_J, jac_T) - tf.matmul(jac, tf.transpose(dot_J)), tf.transpose(tilde_w))
    w_dot = tf.linalg.lstsq(wdot_lhs, wdot_rhs, fast=False)
    w_dot = tf.matmul((Id - tf.einsum('i,j->ij', tilde_w[0], tf.transpose(tilde_w[0]))), w_dot)
    # equation for w:
    x_dot = tf.transpose(tilde_w)
    #
    return tf.transpose(tf.concat([x_dot, w_dot, alpha_dot], axis=0))[0]




#def _naive_KL_ode(t, y, reference, flow, prior_flow):
#    """
#    Solve naively the dynamical equation for KL decomposition in abstract space.
#    """
#    # preprocess:
#    x = tf.convert_to_tensor([tf.cast(y, tf.float32)])
#    # compute metrics:
#    metric = flow.metric(x)[0]
#    prior_metric = prior_flow.metric(x)[0]
#    # compute KL decomposition:
#    eig, eigv = tf_KL_decomposition(metric, prior_metric)
#    # normalize to one to project and select direction:
#    temp = tf.matmul(tf.transpose(eigv), tf.transpose([reference]))[:, 0] / tf.linalg.norm(eigv, axis=0)
#    idx = tf.math.argmax(tf.abs(temp))
#    w = tf.math.sign(temp[idx]) * eigv[:, idx]
#    # normalize affine parameter:
#    s = tf.math.sqrt(tf.tensordot(w, tf.tensordot(metric, w, 1), 1))
#    #
#    return tf.convert_to_tensor([w / s])


def _naive_KL_ode(t, y, reference, flow, prior_flow):
    """
    Solve naively the dynamical equation for KL decomposition in abstract space.

    :param t: integration time.
    :param y: current position in abstract space.
    :param reference: reference eigenvector used to select the branch.
    :param flow: flow defining the target metric.
    :param prior_flow: flow defining the prior metric.
    :returns: normalized direction for the KL mode at the current step.
    """
    # preprocess:
    x = flow.cast([y])
    # compute metrics:
    metric = flow.metric(x)[0]
    prior_metric = prior_flow.metric(x)[0]
    # compute KL decomposition:
    eig, eigv = tf_KL_decomposition(metric, prior_metric)
    # normalize to one to project and select direction:
    #temp = tf.linalg.matvec(tf.matmul(tf.transpose(eigv), metric), reference) - tf.tensordot(reference, tf.tensordot(metric, reference, 1), 1)
    #idx = tf.math.argmin(tf.abs(temp))
    #temp_2 = tf.tensordot(eigv[:, idx], reference, 1)
    #w = tf.math.sign(temp_2) * eigv[:, idx]
    #
    temp = tf.matmul(tf.transpose(eigv), tf.transpose([reference]))[:, 0] / tf.linalg.norm(eigv, axis=0) / tf.linalg.norm(reference)
    idx = tf.math.argmax(tf.abs(temp))
    w = tf.math.sign(temp[idx]) * eigv[:, idx]
    # normalize affine parameter:
    s = tf.math.sqrt(tf.tensordot(w, tf.tensordot(metric, w, 1), 1))
    #
    return tf.convert_to_tensor([w / s])


def solve_KL_ode(flow, prior_flow, y0, n, length=1.5, side='both', integrator_options=None, num_points=100, **kwargs):
    """
    Solve the KL eigenmode ODE in abstract space.

    :param flow: flow representing the target distribution.
    :param prior_flow: flow representing the prior distribution.
    :param y0: starting point in abstract coordinates.
    :param n: index of the KL eigenmode to track.
    :param length: integration length for each direction.
    :param side: which direction to integrate; ``'+'``, ``'-'`` or ``'both'``.
    :param integrator_options: optional options forwarded to ``scipy.integrate.ode``.
    :param num_points: number of output points per direction.
    :returns: solution times, trajectory, and velocity along the path.
    """
    y0_np = np.asarray(y0, dtype=np.float64)
    # define solution points:
    solution_times = tf.linspace(0., length, num_points)
    # compute initial KL decomposition:
    x = flow.cast([y0_np])
    metric = flow.metric(x)[0]
    prior_metric = prior_flow.metric(x)[0]
    # compute KL decomposition:
    eig, eigv = tf_KL_decomposition(metric, prior_metric)
    # solve forward:
    if side == '+' or side == 'both':
        # initialize solution:
        temp_sol_1 = np.zeros((num_points-1, flow.num_params))
        temp_sol_dot_1 = np.zeros((num_points-1, flow.num_params))
        # initialize forward integration:
        solver = scipy.integrate.ode(_naive_KL_ode)
        if integrator_options is not None:
            solver.set_integrator(**integrator_options)
        solver.set_initial_value(y0_np, 0.)
        #reference = eigv[:, n] / tf.norm(eigv[:, n])
        reference = eigv[:, n]
        yt = y0_np.copy()
        yprime = eigv[:, n]
        # do the time steps:
        for ind, t in enumerate(solution_times[1:]):
            # set the reference:
            solver.set_f_params(reference, flow, prior_flow)
            # advance solver:
            try:
                yt = solver.integrate(t)
                yprime = _naive_KL_ode(t, yt, reference, flow, prior_flow)
            except:
                pass
            # update reference:
            # reference = yprime[0] / tf.norm(yprime[0])
            reference = yprime[0]
            # save out:
            temp_sol_1[ind] = yt.copy()
            temp_sol_dot_1[ind] = yprime.numpy().copy()
        # return if needed:
        if side == '+':
            traj = np.concatenate((x.numpy(), temp_sol_1))
            vel = np.concatenate(([eigv[:, n].numpy()], temp_sol_dot_1))
            return solution_times, traj, vel
    # solve backward:
    if side == '-' or side == 'both':
        # initialize solution:
        temp_sol_2 = np.zeros((num_points-1, flow.num_params))
        temp_sol_dot_2 = np.zeros((num_points-1, flow.num_params))
        # initialize backward integration:
        solver = scipy.integrate.ode(_naive_KL_ode)
        if integrator_options is not None:
            solver.set_integrator(**integrator_options)
        solver.set_initial_value(y0_np, 0.)
        # reference = - eigv[:, n] / tf.norm(eigv[:, n])
        reference = - eigv[:, n]
        yt = y0_np.copy()
        yprime = reference
        for ind, t in enumerate(solution_times[1:]):
            # set the reference:
            solver.set_f_params(reference, flow, prior_flow)
            # advance solver:
            try:
                yt = solver.integrate(t)
                yprime = _naive_KL_ode(t, yt, reference, flow, prior_flow)
            except:
                pass
            # update reference:
            # reference = yprime[0] / tf.norm(yprime[0])
            reference = yprime[0]
            # save out:
            temp_sol_2[ind] = yt.copy()
            temp_sol_dot_2[ind] = yprime.numpy().copy()
        # return if needed:
        if side == '-':
            traj = np.concatenate((temp_sol_2[::-1], x.numpy()))
            vel = np.concatenate((-temp_sol_dot_2[::-1], [eigv[:, n].numpy()]))
            return -solution_times, traj, vel
    # patch solutions:
    times = np.concatenate((-solution_times[::-1], solution_times[1:]))
    traj = np.concatenate((temp_sol_2[::-1], x.numpy(), temp_sol_1))
    vel = np.concatenate((-temp_sol_dot_2[::-1], [eigv[:, n].numpy()], temp_sol_dot_1))
    #
    return times, traj, vel



"""Tests for KDE-based MCMC tension estimators."""

#########################################################################################################
# Imports

import builtins
import importlib
import os
import types
import unittest
from unittest.mock import patch

import numpy as np
from getdist import MCSamples, loadMCSamples

import tensiometer.mcmc_tension.kde as mt
import tensiometer.mcmc_tension.param_diff as pd
import tensiometer.utilities.stats_utilities as stut

#########################################################################################################
# KDE shift tests


class TestMcmcKdeShift(unittest.TestCase):

    """MCMC KDE shift test suite."""
    def setUp(self):
        """Set up test fixtures."""
        self.here = os.path.dirname(os.path.abspath(__file__))
        self.chain_1 = loadMCSamples(self.here + "/../../test_chains/DES")
        self.chain_2 = loadMCSamples(self.here + "/../../test_chains/Planck18TTTEEE")
        self.chain_12 = loadMCSamples(self.here + "/../../test_chains/Planck18TTTEEE_DES")
        self.chain_prior = loadMCSamples(self.here + "/../../test_chains/prior")
        # keep joblib sequential for test environments that block semaphores
        mt.n_threads = 1
        # patch logsumexp to avoid SciPy array-api quirks in this environment
        self._orig_logsumexp = mt.scipy.special.logsumexp
        mt.scipy.special.logsumexp = lambda a, b=None, axis=None, return_sign=False, xp=None: np.log(
            np.sum(np.exp(a)*(1 if b is None else b))
        )
        # Disable numba JIT during tests to avoid dispatcher compilation in small fixtures.
        self._orig_jit = mt.jit
        mt.jit = lambda *args, **kwargs: (lambda f: f)
        # thin the chain:
        self.chain_1.getConvergeTests()
        self.chain_2.getConvergeTests()
        self.chain_12.getConvergeTests()
        self.chain_prior.getConvergeTests()
        self.chain_1.weighted_thin(int(self.chain_1.indep_thin))
        self.chain_2.weighted_thin(int(self.chain_2.indep_thin))
        self.chain_12.weighted_thin(int(self.chain_12.indep_thin))
        self.chain_prior.weighted_thin(int(self.chain_prior.indep_thin))
        # Keep numba helpers running in pure Python to avoid JIT compile errors in tests.
        self._orig_mise1d = mt._mise1d_optimizer
        self._orig_mise1d_jac = mt._mise1d_optimizer_jac
        self._orig_mise = mt._mise_optimizer
        self._orig_ell_helper = mt._ell_helper
        if hasattr(mt._mise1d_optimizer, "py_func"):
            mt._mise1d_optimizer = mt._mise1d_optimizer.py_func
        if hasattr(mt._mise1d_optimizer_jac, "py_func"):
            mt._mise1d_optimizer_jac = mt._mise1d_optimizer_jac.py_func
        if hasattr(mt._mise_optimizer, "py_func"):
            mt._mise_optimizer = mt._mise_optimizer.py_func
        if hasattr(mt._ell_helper, "py_func"):
            mt._ell_helper = mt._ell_helper.py_func
        self.diff_chain = pd.parameter_diff_chain(self.chain_1, self.chain_2, boost=1)

    def tearDown(self):
        """Clean up test fixtures."""
        mt.scipy.special.logsumexp = self._orig_logsumexp
        mt.jit = self._orig_jit
        mt._mise1d_optimizer = self._orig_mise1d
        mt._mise1d_optimizer_jac = self._orig_mise1d_jac
        mt._mise_optimizer = self._orig_mise
        mt._ell_helper = self._orig_ell_helper

    def test_kde_shift(self):
        """Test KDE shift consistency across methods."""
        res_1 = mt.kde_parameter_shift(self.diff_chain,
                                       method='brute_force',
                                       scale=0.5)
        res_2 = mt.kde_parameter_shift(self.diff_chain,
                                       method='neighbor_elimination',
                                       scale=0.5)
        print(res_1, res_2)
        assert np.allclose(res_1, res_2)
        res_3 = mt.kde_parameter_shift(self.diff_chain,
                                       method='brute_force',
                                       feedback=2,
                                       scale=0.5)
        assert np.allclose(res_1, res_3)

    def test_helper_branches(self):
        """Test helper branches and error paths."""
        samples = self.diff_chain.samples[:50, :2]
        weights = self.diff_chain.weights[:50]
        bw = mt.UCV_SP_bandwidth(samples, weights, near=2, feedback=0)
        self.assertTrue(hasattr(bw, "x"))
        res = mt.kde_parameter_shift(self.diff_chain,
                                     method='neighbor_elimination',
                                     scale=0.5,
                                     zero_prob=-1.0,
                                     num_samples=100,
                                     feedback=0)
        self.assertTrue(np.isfinite(res[0]))
        coords = np.random.rand(10, 2)
        dist_weights = np.ones(10)
        mt._gauss_ballkde_logpdf(coords[0], coords, weights[:10], dist_weights)
        ell_weights = np.array([np.eye(2) for _ in range(10)])
        mt._gauss_ellkde_logpdf(coords[0], coords, weights[:10], ell_weights)
        with self.assertRaises(ValueError):
            mt.kde_parameter_shift(self.diff_chain, method='invalid')
        res_4 = mt.kde_parameter_shift(self.diff_chain,
                                       method='neighbor_elimination',
                                       feedback=2,
                                       scale=0.5)
        self.assertTrue(np.isfinite(res_4[0]))
        param_names = ['delta_omegam', 'delta_sigma8']
        res_1 = mt.kde_parameter_shift(self.diff_chain,
                                       param_names=param_names,
                                       method='brute_force',
                                       scale=0.5)
        res_2 = mt.kde_parameter_shift(self.diff_chain,
                                       param_names=param_names,
                                       method='neighbor_elimination',
                                       scale=0.5)
        print(res_1, res_2)
        assert np.allclose(res_1, res_2)

    def test_band(self):
        """Test KDE bandwidth selectors."""
        n, d = self.diff_chain.samples.shape
        weights = self.diff_chain.weights
        wtot = np.sum(weights)
        neff = wtot**2 / np.sum(weights**2)
        mt.Scotts_bandwidth(d, neff)
        mt.AMISE_bandwidth(d, neff)
        mt.MAX_bandwidth(d, neff)
        mt.MISE_bandwidth_1d(d, neff)
        mt.MISE_bandwidth(d, neff)
        white_samples = stut.whiten_samples(self.diff_chain.samples, weights)
        mt.UCV_bandwidth(weights, white_samples, mode='1d', feedback=1)
        mt.UCV_SP_bandwidth(white_samples, weights, near=1, near_max=20, feedback=1)

    def test_fft_shift(self):
        """Test FFT KDE shift helpers."""
        param_names = ['delta_sigma8']
        mt.kde_parameter_shift_1D_fft(self.diff_chain, param_names=param_names, feedback=2)
        param_names = ['delta_omegam', 'delta_sigma8']
        mt.kde_parameter_shift_2D_fft(self.diff_chain, param_names=param_names, feedback=2)

    def test_ball_kde(self):
        """Test BALL and ELL KDE distance scales."""
        res_1 = mt.kde_parameter_shift(self.diff_chain,
                                       method='brute_force',
                                       feedback=2,
                                       scale='BALL')
        res_2 = mt.kde_parameter_shift(self.diff_chain,
                                       method='neighbor_elimination',
                                       feedback=2,
                                       scale='BALL')
        assert np.allclose(res_1, res_2)

        res_1 = mt.kde_parameter_shift(self.diff_chain,
                                       method='brute_force',
                                       feedback=2,
                                       scale='ELL')
        res_2 = mt.kde_parameter_shift(self.diff_chain,
                                       method='neighbor_elimination',
                                       feedback=2,
                                       scale='ELL')
        assert np.allclose(res_1, res_2)

#########################################################################################################
# Helper utilities


def make_chain_additional(dim=2, n=4):
    """Build a small MCSamples chain with simple names and weights.

    :param dim: number of parameters.
    :param n: number of samples.
    :returns: chain with ``n`` samples and ``dim`` parameters.
    """
    rng = np.linspace(0, 1, n * dim).reshape(n, dim)
    samples = rng + np.eye(n, dim, k=0) * 0.05
    weights = np.ones(n)
    names = [f"p{i}" for i in range(dim)]
    return MCSamples(samples=samples, weights=weights, names=names,
                     labels=names, sampler="uncorrelated", ignore_rows=0)


def make_chain_branch(dim=2, n=4):
    """Build a deterministic MCSamples chain.

    :param dim: number of parameters.
    :param n: number of samples.
    :returns: chain with evenly spaced samples.
    """
    data = np.linspace(0, 1, n * dim).reshape(n, dim)
    weights = np.ones(n)
    names = [f"p{i}" for i in range(dim)]
    chain = MCSamples(samples=data, weights=weights, names=names, labels=names,
                      sampler="uncorrelated", ignore_rows=0)
    return chain


class DummyParallel:
    """Replacement for joblib.Parallel that runs sequentially."""

    def __init__(self, *args, **kwargs):
        """Init."""
        pass

    def __enter__(self):
        """Enter."""
        return self

    def __exit__(self, *args):
        """Exit."""
        return False

    def __call__(self, tasks):
        """Call."""
        return np.array([t() if callable(t) else t for t in tasks])

#########################################################################################################
# Additional KDE tests


class TestKdeAdditional(unittest.TestCase):
    """KDE additional test suite."""
    def setUp(self):
        """Set up test fixtures."""
        # Disable numba to keep helpers in pure Python and reload module accordingly.
        self._orig_disable_jit = os.environ.get("NUMBA_DISABLE_JIT")
        os.environ["NUMBA_DISABLE_JIT"] = "1"
        globals()["mt"] = importlib.reload(mt)
        mt.n_threads = 1
        self._orig_logsumexp = mt.scipy.special.logsumexp
        mt.scipy.special.logsumexp = lambda a, b=None, axis=None, return_sign=False, xp=None: np.log(
            np.sum(np.exp(a)*(1 if b is None else b))
        )
        # Replace numba.jit with a no-op to avoid generating Dispatchers at runtime.
        self._orig_jit = mt.jit
        mt.jit = lambda *args, **kwargs: (lambda f: f)
        # Force numba-jitted helpers to run in pure Python to avoid compile errors.
        self._orig_mise1d = mt._mise1d_optimizer
        self._orig_mise1d_jac = mt._mise1d_optimizer_jac
        self._orig_mise = mt._mise_optimizer
        self._orig_ell_helper = mt._ell_helper
        if hasattr(mt._mise1d_optimizer, "py_func"):
            mt._mise1d_optimizer = mt._mise1d_optimizer.py_func
        if hasattr(mt._mise1d_optimizer_jac, "py_func"):
            mt._mise1d_optimizer_jac = mt._mise1d_optimizer_jac.py_func
        if hasattr(mt._mise_optimizer, "py_func"):
            mt._mise_optimizer = mt._mise_optimizer.py_func
        if hasattr(mt._ell_helper, "py_func"):
            mt._ell_helper = mt._ell_helper.py_func

    def tearDown(self):
        """Clean up test fixtures."""
        mt.scipy.special.logsumexp = self._orig_logsumexp
        mt._mise1d_optimizer = self._orig_mise1d
        mt._mise1d_optimizer_jac = self._orig_mise1d_jac
        mt._mise_optimizer = self._orig_mise
        mt._ell_helper = self._orig_ell_helper
        mt.jit = self._orig_jit
        if self._orig_disable_jit is None:
            os.environ.pop("NUMBA_DISABLE_JIT", None)
        else:
            os.environ["NUMBA_DISABLE_JIT"] = self._orig_disable_jit
        globals()["mt"] = importlib.reload(mt)

    def test_force_disable_jit_pop(self):
        """Force NUMBA_DISABLE_JIT cleanup path."""
        self._orig_disable_jit = None
        os.environ["NUMBA_DISABLE_JIT"] = "1"

    def test_import_branches_and_env(self):
        """Test Import branches and env."""
        # Force ImportError for simpson to hit fallback and set env for threads.
        path = os.path.join(os.path.dirname(mt.__file__), "kde.py")
        real_import = builtins.__import__
        seen = {"first": True}

        def fake_import(name, globals=None, locals=None, fromlist=(), level=0):
            """Fake import."""
            if name == "scipy.integrate" and "simpson" in fromlist:
                if seen["first"]:
                    seen["first"] = False
                    raise ImportError("force fallback")
            if name == "scipy.integrate" and "simps" in fromlist:
                dummy = types.SimpleNamespace(simps=lambda x: x)
                return dummy
            return real_import(name, globals, locals, fromlist, level)

        with patch.dict(os.environ, {"OMP_NUM_THREADS": "2"}):
            with patch("builtins.__import__", side_effect=fake_import):
                importlib.reload(mt)
                self.assertEqual(mt.n_threads, 2)
        importlib.reload(mt)

    def test_bandwidth_helpers(self):
        """Test Bandwidth helpers."""
        mt._mise1d_optimizer = types.SimpleNamespace(py_func=mt._mise1d_optimizer)
        mt._mise1d_optimizer_jac = types.SimpleNamespace(py_func=mt._mise1d_optimizer_jac)
        mt._mise_optimizer = types.SimpleNamespace(py_func=mt._mise_optimizer)
        # Directly exercise numba helpers and MISE options.
        # use the python implementations to avoid JIT compilation issues
        if hasattr(mt._mise1d_optimizer, "py_func"):
            mt._mise1d_optimizer = mt._mise1d_optimizer.py_func
        if hasattr(mt._mise1d_optimizer_jac, "py_func"):
            mt._mise1d_optimizer_jac = mt._mise1d_optimizer_jac.py_func
        if hasattr(mt._mise_optimizer, "py_func"):
            mt._mise_optimizer = mt._mise_optimizer.py_func
        mt._mise1d_optimizer(0.5, 1, 10)
        mt._mise1d_optimizer_jac(0.5, 1, 10)
        mt._mise_optimizer(np.eye(1), 1, 5)
        res = mt.MISE_bandwidth_1d(1, 5, alpha0=0.2)
        self.assertTrue(np.all(np.isfinite(res)))
        res2 = mt.MISE_bandwidth(2, 6, alpha0=np.eye(2) * 0.3, feedback=3)
        self.assertEqual(res2.shape, (2, 2))

    def test_ucv_modes(self):
        """Test Ucv modes."""
        samples = np.array([[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]])
        weights = np.ones(3)
        res1 = mt.UCV_bandwidth(weights, samples, mode="1d", n_nearest=3)
        self.assertEqual(res1.shape, (2, 2))
        res2 = mt.UCV_bandwidth(weights, samples, mode="diag", n_nearest=3)
        self.assertEqual(res2.shape, (2, 2))
        res3 = mt.UCV_bandwidth(weights, samples, mode="full", n_nearest=3,
                                bounds=np.array([[0.001, 10], [0.001, 10], [0.001, 10]]))
        self.assertEqual(res3.shape, (2, 2))
        res_sp = mt.UCV_SP_bandwidth(samples, weights, near=1, near_max=3)
        self.assertTrue(np.isfinite(res_sp.x))

    def test_mise_bandwidth_default_bounds(self):
        """Test MISE_bandwidth default bounds path."""
        opt = types.SimpleNamespace(success=True, x=mt.stutils.PDM_to_vector(np.eye(2)))
        with patch.object(mt.scipy.optimize, "minimize", return_value=opt):
            res = mt.MISE_bandwidth(2, 5, feedback=0)
        self.assertEqual(res.shape, (2, 2))

    def test_ucv_bandwidth_default_alpha0_1d(self):
        """Test UCV_bandwidth default alpha0 in 1D mode."""
        samples = np.array([[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]])
        weights = np.ones(3)
        opt = types.SimpleNamespace(success=True, x=np.array([0.1]))
        with patch.object(mt.scipy.optimize, "minimize", return_value=opt):
            res = mt.UCV_bandwidth(weights, samples, mode="1d", feedback=0)
        self.assertEqual(res.shape, (2, 2))

    def test_ucv_bandwidth_full_feedback(self):
        """Test UCV_bandwidth full mode feedback."""
        samples = np.array([[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]])
        weights = np.ones(3)
        opt = types.SimpleNamespace(success=True, x=mt.stutils.PDM_to_vector(np.eye(2)))
        with patch.object(mt.scipy.optimize, "minimize", return_value=opt):
            res = mt.UCV_bandwidth(weights, samples, mode="full", feedback=3)
        self.assertEqual(res.shape, (2, 2))

    def test_brute_force_and_neighbor_helpers(self):
        """Test Brute force and neighbor helpers."""
        white_samples = np.array([[0.0, 0.0], [0.1, 0.1]])
        weights = np.array([1.0, 1.0])
        with patch("tensiometer.mcmc_tension.kde.joblib.Parallel", DummyParallel), \
                patch("tensiometer.mcmc_tension.kde.joblib.delayed",
                      lambda f: (lambda *a, **k: lambda: f(*a, **k))):
            mt._brute_force_kde_param_shift(white_samples, weights, -10, 2, 0)
            # BALL variant
            mt._brute_force_kde_param_shift(white_samples, weights, -10, 2, 0,
                                            distance_weights=np.ones(2))
            # ELL variant
            ell = np.array([np.eye(2), np.eye(2)])
            mt._brute_force_kde_param_shift(white_samples, weights, -10, 2, 0,
                                            distance_weights=ell, weights_norm=weights)
            mt._neighbor_parameter_shift(white_samples, weights, -10, 2, 0,
                                         stable_cycle=1, smallest_improvement=0,
                                         chunk_size=2)

    def test_kde_parameter_shift_scales_and_errors(self):
        """Test KDE parameter shift scales and errors."""
        chain = make_chain_additional()
        bad_scale = np.ones((3, 3))
        with patch("tensiometer.mcmc_tension.kde.joblib.Parallel", DummyParallel), \
                patch("tensiometer.mcmc_tension.kde.joblib.delayed",
                      lambda f: (lambda *a, **k: lambda: f(*a, **k))):
            for scale in ["MISE", "AMISE", "MAX", 0.2]:
                mt.kde_parameter_shift(chain, scale=scale, feedback=0, num_samples=2)
            mt.kde_parameter_shift(chain, scale="BALL", feedback=0, near=1)
            mt.kde_parameter_shift(chain, scale="ELL", feedback=0)
            with self.assertRaises(ValueError):
                mt.kde_parameter_shift(chain, scale=bad_scale)
            with self.assertRaises(ValueError):
                mt.kde_parameter_shift(chain, method="unknown")
            # zero_prob and num_samples override branches
            mt.kde_parameter_shift(chain, scale=0.1, feedback=0,
                                   zero_prob=-1.0, num_samples=1)

    def test_kde_param_names_and_scale_errors(self):
        """Test parameter name and scale validation errors."""
        chain = make_chain_additional()
        with self.assertRaises(ValueError):
            mt.kde_parameter_shift(chain, param_names=["missing"])
        with self.assertRaises(ValueError):
            mt.kde_parameter_shift(chain, scale={"bad": 1})

    def test_kde_scale_matrix_feedback(self):
        """Test non-diagonal scale feedback in KDE."""
        chain = make_chain_additional()
        scale = np.array([[1.0, 0.2], [0.2, 1.0]])
        with patch.object(mt, "_neighbor_parameter_shift", return_value=0):
            mt.kde_parameter_shift(chain, scale=scale, feedback=1)

    def test_kde_ell_scale_feedback(self):
        """Test ELL scale feedback and setup."""
        chain = make_chain_additional()
        dets = np.ones(chain.samples.shape[0])
        mats = [np.eye(chain.samples.shape[1]) for _ in range(chain.samples.shape[0])]
        with patch.object(mt, "_ell_helper", return_value=(dets, mats)), \
                patch.object(mt, "_neighbor_parameter_shift", return_value=0):
            mt.kde_parameter_shift(chain, scale="ELL", feedback=1)

    def test_fft_and_optimize_bandwidth_errors(self):
        """Test FFT and optimize bandwidth errors."""
        chain = make_chain_additional(dim=1)
        # wrong parameter count for 1D and 2D FFT
        with self.assertRaises(ValueError):
            mt.kde_parameter_shift_1D_fft(chain, param_names=["p0", "extra"])
        with self.assertRaises(ValueError):
            mt.kde_parameter_shift_2D_fft(chain, param_names=["p0"])
        # prior mismatch
        prior = make_chain_additional(dim=1)
        with self.assertRaises(ValueError):
            mt.kde_parameter_shift_1D_fft(chain, prior_diff_chain=prior, param_names=["missing"])
        # valid 1D call
        mt.kde_parameter_shift_1D_fft(chain, feedback=0)
        # Optimize bandwidth bad parameter
        with self.assertRaises(ValueError):
            mt.OptimizeBandwidth_1D(chain, param_names=["missing"])
        mt.OptimizeBandwidth_1D(chain)

    def test_fft_prior_checks_and_probability_path(self):
        """Test FFT prior checks and probability path."""
        # 2D FFT path with prior and no samples above zero
        chain = make_chain_additional(dim=2)
        prior = make_chain_additional(dim=2)
        prior.weights *= 0.5
        mt.kde_parameter_shift_2D_fft(chain, prior_diff_chain=prior, feedback=0, nbins=16)

#########################################################################################################
# Branch coverage tests


class TestKdeBranchCoverage(unittest.TestCase):
    """KDE branch coverage test suite."""
    def setUp(self):
        """Set up test fixtures."""
        # Disable numba so jitted helpers execute in Python and count for coverage.
        self._orig_numba = os.environ.get("NUMBA_DISABLE_JIT")
        os.environ["NUMBA_DISABLE_JIT"] = "1"
        self.mt = importlib.reload(mt)
        # Make jit a no-op so inner helpers stay as plain Python functions.
        self._orig_jit = self.mt.jit
        self.mt.jit = lambda *args, **kwargs: (lambda f: f)
        # Keep python implementations of jitted helpers to avoid compile errors.
        self._orig_mise1d = self.mt._mise1d_optimizer
        self._orig_mise1d_jac = self.mt._mise1d_optimizer_jac
        self._orig_mise = self.mt._mise_optimizer
        self._orig_ell_helper = self.mt._ell_helper
        self._orig_ucv_brute = self.mt._UCV_optimizer_brute_force
        if hasattr(self.mt._mise1d_optimizer, "py_func"):
            self.mt._mise1d_optimizer = self.mt._mise1d_optimizer.py_func
        if hasattr(self.mt._mise1d_optimizer_jac, "py_func"):
            self.mt._mise1d_optimizer_jac = self.mt._mise1d_optimizer_jac.py_func
        if hasattr(self.mt._mise_optimizer, "py_func"):
            self.mt._mise_optimizer = self.mt._mise_optimizer.py_func
        if hasattr(self.mt._ell_helper, "py_func"):
            self.mt._ell_helper = self.mt._ell_helper.py_func
        if hasattr(self.mt._UCV_optimizer_brute_force, "py_func"):
            self.mt._UCV_optimizer_brute_force = self.mt._UCV_optimizer_brute_force.py_func
        # Patch logsumexp to avoid SciPy array-api quirks in this environment.
        self._orig_logsumexp = self.mt.scipy.special.logsumexp
        self.mt.scipy.special.logsumexp = lambda a, b=None, axis=None, return_sign=False, xp=None: np.log(
            np.sum(np.exp(a)*(1 if b is None else b))
        )

    def tearDown(self):
        """Clean up test fixtures."""
        if self._orig_numba is None:
            os.environ.pop("NUMBA_DISABLE_JIT", None)
        else:
            os.environ["NUMBA_DISABLE_JIT"] = self._orig_numba
        importlib.reload(mt)
        mt.scipy.special.logsumexp = self._orig_logsumexp
        mt.jit = self._orig_jit
        mt._mise1d_optimizer = self._orig_mise1d
        mt._mise1d_optimizer_jac = self._orig_mise1d_jac
        mt._mise_optimizer = self._orig_mise
        mt._ell_helper = self._orig_ell_helper
        mt._UCV_optimizer_brute_force = self._orig_ucv_brute

    def test_numba_disable_env_cleanup(self):
        """Exercise NUMBA_DISABLE_JIT cleanup branch."""
        self._orig_numba = None
        os.environ["NUMBA_DISABLE_JIT"] = "1"

    def test_numba_disable_env_restore(self):
        """Exercise NUMBA_DISABLE_JIT restore branch."""
        self._orig_numba = "0"
        os.environ["NUMBA_DISABLE_JIT"] = "1"

    def test_numba_helpers_and_mise_failures(self):
        """Test Numba helpers and mise failures."""
        # Directly exercise the jitted helpers via Python execution.
        self.mt._mise1d_optimizer(0.5, 1, 10)
        self.mt._mise1d_optimizer_jac(0.5, 1, 10)
        self.mt._mise_optimizer(np.eye(1), 1, 5)
        weights = np.array([1.0, 1.0])
        samples = np.array([[0.0], [0.1]])
        self.mt._UCV_optimizer_brute_force(np.eye(1), weights, samples)
        self.mt._ell_helper(np.array([[0, 1]]), np.array([[[0.0]], [[0.1]]])[:, :, 0], 1)
        # Force optimizer failures to hit print-feedback branches.
        bad_opt = types.SimpleNamespace(success=False, x=np.array([0.1]))
        bad_opt2 = types.SimpleNamespace(success=False, x=np.array([0.1, 0.1, 0.1]))
        with patch.object(self.mt.scipy.optimize, "minimize", return_value=bad_opt):
            self.mt.MISE_bandwidth_1d(1, 5, alpha0=0.2)
        with patch.object(self.mt.scipy.optimize, "minimize", return_value=bad_opt2):
            self.mt.MISE_bandwidth(2, 6)

    def test_ucv_bandwidth_modes_and_feedback(self):
        """Test Ucv bandwidth modes and feedback."""
        weights = np.ones(3)
        samples = np.array([[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]])
        # Make optimizer fail to trigger logging and default bounds handling.
        opt_fail = types.SimpleNamespace(success=False, x=np.array([0.2]))
        opt_fail_diag = types.SimpleNamespace(success=False, x=np.array([0.2, 0.2]))
        opt_fail_full = types.SimpleNamespace(success=False, x=np.array([0.2, 0.2, 0.2]))
        with patch.object(self.mt.scipy.optimize, "minimize", side_effect=[opt_fail, opt_fail_diag, opt_fail_full]):
            self.mt.UCV_bandwidth(weights, samples, mode="1d", n_nearest=3, feedback=1)
            self.mt.UCV_bandwidth(weights, samples, mode="diag", n_nearest=3, feedback=1)
            self.mt.UCV_bandwidth(weights, samples, mode="full", n_nearest=3, feedback=3)

    def test_mise_bandwidth_with_bounds(self):
        """Test MISE bandwidth with explicit bounds."""
        opt = types.SimpleNamespace(success=True, x=self.mt.stutils.PDM_to_vector(np.eye(2)))
        bounds = np.array([[0.1, 1.0], [0.1, 1.0], [0.1, 1.0]])
        with patch.object(self.mt.scipy.optimize, "minimize", return_value=opt):
            res = self.mt.MISE_bandwidth(2, 5, bounds=bounds, feedback=0)
        self.assertEqual(res.shape, (2, 2))

    def test_ucv_bandwidth_alpha0_supplied(self):
        """Test UCV bandwidth with explicit alpha0."""
        weights = np.ones(3)
        samples = np.array([[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]])
        opt = types.SimpleNamespace(success=True, x=np.array([0.2, 0.2]))
        with patch.object(self.mt.scipy.optimize, "minimize", return_value=opt):
            res = self.mt.UCV_bandwidth(weights, samples, mode="diag", alpha0=np.eye(2), feedback=0)
        self.assertEqual(res.shape, (2, 2))

    def test_ucv_bandwidth_invalid_mode(self):
        """Test UCV bandwidth invalid mode error."""
        weights = np.ones(3)
        samples = np.array([[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]])
        with self.assertRaises(ValueError):
            self.mt.UCV_bandwidth(weights, samples, mode="bad")

    def test_optimize_bandwidth_with_param_names(self):
        """Test OptimizeBandwidth_1D with explicit parameter names."""
        chain = make_chain_additional()
        res = self.mt.OptimizeBandwidth_1D(chain, param_names=["p0"])
        self.assertEqual(res.shape, (1,))

    def test_fft_numeric_scale_branches(self):
        """Test FFT branches with numeric scale inputs."""
        chain = make_chain_branch(dim=1)

        class FakeDensity1D:
            """Fake Density1 D test suite."""
            def __init__(self):
                """Init."""
                self.P = np.array([1.0, 0.5])
                self.x = np.array([0.0, 1.0])

            def _initSpline(self):  # pragma: no cover - trivial
                """Init Spline."""
                return None

            def Prob(self, arr):
                """Prob."""
                arr = np.atleast_1d(arr)
                return np.full_like(arr, 0.5, dtype=float)

        with patch.object(chain, "get1DDensity", return_value=FakeDensity1D()):
            self.mt.kde_parameter_shift_1D_fft(chain, param_names=["p0"], feedback=0, nbins=8, scale=0.5)

        chain2d = make_chain_branch(dim=2)

        class DummySpl:
            """Dummy Spl test suite."""
            def __call__(self, xs, ys):
                """Call."""
                return np.array([[0.1]])

            def ev(self, x, y):
                """Ev."""
                return np.zeros_like(x, dtype=float)

        class FakeDensity2D:
            """Fake Density2 D test suite."""
            def __init__(self):
                """Init."""
                self.P = np.full((2, 2), 0.1)
                self.x = np.array([0.0, 1.0])
                self.y = np.array([0.0, 1.0])
                self.spl = DummySpl()

            def _initSpline(self):  # pragma: no cover - trivial
                """Init Spline."""
                return None

        with patch.object(chain2d, "get2DDensity", return_value=FakeDensity2D()):
            self.mt.kde_parameter_shift_2D_fft(chain2d, feedback=0, nbins=8, scale=0.5)

    def test_kde_ell_feedback_and_logprob(self):
        """Test ELL scale feedback and log-probability setup."""
        chain = make_chain_additional()
        dets = np.ones(chain.samples.shape[0])
        mats = [np.eye(chain.samples.shape[1]) for _ in range(chain.samples.shape[0])]
        def _fake_logpdf(point, samples, weights, distance_weights):
            """Fake logpdf."""
            return -1.0
        with patch.object(self.mt, "_ell_helper", return_value=(dets, mats)), \
                patch.object(self.mt, "_neighbor_parameter_shift", return_value=0), \
                patch.object(self.mt, "_gauss_ellkde_logpdf", side_effect=_fake_logpdf):
            res = self.mt.kde_parameter_shift(chain, scale="ELL", feedback=1)
        self.assertTrue(np.isfinite(res[0]))

    def test_neighbor_elimination_feedback_noise(self):
        """Test Neighbor elimination feedback noise."""
        white_samples = np.array([[0.0, 0.0], [0.05, 0.05], [0.1, 0.1], [0.15, 0.15]])
        weights = np.ones(4)
        # Run with verbose feedback and small chunk to hit logging branches.
        with patch("tensiometer.mcmc_tension.kde.joblib.Parallel", DummyParallel), \
                patch("tensiometer.mcmc_tension.kde.joblib.delayed",
                      lambda f: (lambda *a, **k: lambda: f(*a, **k))):
            self.mt._neighbor_parameter_shift(white_samples, weights, -10, 4, feedback=3,
                                              stable_cycle=1, chunk_size=2, smallest_improvement=0.0)

    def test_fft_branches_with_fake_densities(self):
        """Test FFT branches with fake densities."""
        chain = make_chain_branch(dim=1)

        class FakeDensity1D:
            """Fake Density1 D test suite."""
            def __init__(self):
                """Init."""
                self.P = np.array([1.0, 0.5])
                self.x = np.array([0.0, 1.0])

            def _initSpline(self):  # pragma: no cover - trivial
                """Init Spline."""
                return None

            def Prob(self, arr):
                """Prob."""
                arr = np.atleast_1d(arr)
                return np.full_like(arr, 0.5, dtype=float)

        # No prior, integral branch
        with patch.object(chain, "get1DDensity", return_value=FakeDensity1D()):
            self.mt.kde_parameter_shift_1D_fft(chain, param_names=["p0"], feedback=0, nbins=8)
            self.mt.kde_parameter_shift_1D_fft(chain, param_names=["p0"], feedback=0, nbins=8, scale="MISE")
        # With prior, same integral path
        prior = make_chain_branch(dim=1)
        with patch.object(chain, "get1DDensity", return_value=FakeDensity1D()), \
                patch.object(prior, "get1DDensity", return_value=FakeDensity1D()):
            self.mt.kde_parameter_shift_1D_fft(chain, prior_diff_chain=prior, param_names=["p0"], feedback=0, nbins=8)

        # 2D fake densities to trigger no-sample branches and prior handling
        chain2d = make_chain_branch(dim=2)
        prior2d = make_chain_branch(dim=2)

        class DummySpl:
            """Dummy Spl test suite."""
            def __call__(self, xs, ys):
                """Call."""
                return np.array([[0.1]])

            def ev(self, x, y):
                """Ev."""
                return np.zeros_like(x, dtype=float)

        class FakeDensity2D:
            """Fake Density2 D test suite."""
            def __init__(self):
                """Init."""
                self.P = np.full((2, 2), 0.1)
                self.x = np.array([0.0, 1.0])
                self.y = np.array([0.0, 1.0])
                self.spl = DummySpl()

            def _initSpline(self):  # pragma: no cover - trivial
                """Init Spline."""
                return None

        with patch.object(chain2d, "get2DDensity", return_value=FakeDensity2D()):
            self.mt.kde_parameter_shift_2D_fft(chain2d, feedback=0, nbins=8)
            self.mt.kde_parameter_shift_2D_fft(chain2d, feedback=0, nbins=8, scale="MISE")
        with patch.object(chain2d, "get2DDensity", return_value=FakeDensity2D()), \
                patch.object(prior2d, "get2DDensity", return_value=FakeDensity2D()):
            self.mt.kde_parameter_shift_2D_fft(chain2d, prior_diff_chain=prior2d, feedback=0, nbins=8)
        with self.assertRaises(ValueError):
            self.mt.kde_parameter_shift_2D_fft(chain2d, param_names=["missing"])
        with self.assertRaises(ValueError):
            self.mt.kde_parameter_shift_2D_fft(chain2d, prior_diff_chain=prior, param_names=["p0", "p1"])

    def test_fft_param_validation_branches(self):
        """Test FFT parameter validation branches."""
        chain = make_chain_branch(dim=1)
        prior = make_chain_branch(dim=1)
        prior_mismatch = MCSamples(samples=prior.samples, weights=prior.weights,
                                   names=["q0"], labels=["q0"], sampler="uncorrelated", ignore_rows=0)
        with self.assertRaises(ValueError):
            self.mt.kde_parameter_shift_1D_fft(chain, prior_diff_chain=prior_mismatch, param_names=["p0"])
        chain2d = make_chain_branch(dim=2)
        with self.assertRaises(ValueError):
            self.mt.kde_parameter_shift_1D_fft(chain2d, param_names=["p0", "p1"])

    def test_fft_prior_zero_density_branch(self):
        """Test FFT branch when prior density is zero at origin."""
        chain2d = make_chain_branch(dim=2)
        prior2d = make_chain_branch(dim=2)

        class DummySpl:
            """Dummy spline with zero prior."""
            def __call__(self, xs, ys):
                """Call."""
                return np.array([[0.0]])

            def ev(self, x, y):
                """Evaluate."""
                return np.zeros_like(x, dtype=float)

        class FakeDensity2D:
            """Fake 2D density for zero-prior branch."""
            def __init__(self, zero=False):
                """Init."""
                self.P = np.full((2, 2), 0.1)
                self.x = np.array([0.0, 1.0])
                self.y = np.array([0.0, 1.0])
                self.spl = DummySpl() if zero else DummySpl()

            def _initSpline(self):  # pragma: no cover - trivial
                """Init spline."""
                return None

        with patch.object(chain2d, "get2DDensity", return_value=FakeDensity2D(zero=False)), \
                patch.object(prior2d, "get2DDensity", return_value=FakeDensity2D(zero=True)):
            res = self.mt.kde_parameter_shift_2D_fft(chain2d, prior_diff_chain=prior2d, feedback=0, nbins=8)
            self.assertEqual(res, (1.0, None, None))

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""Tests for analytic flow utilities and wrappers."""

#########################################################################################################
# Imports

import unittest
from unittest.mock import patch

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

from tensiometer.synthetic_probability import analytic_flow as af

#########################################################################################################
# TensorFlow Probability aliases

tfd = tfp.distributions

#########################################################################################################
# Helper distributions


class DummyDistribution:
    """Simple Gaussian-like distribution for exercising analytic_flow."""

    def __init__(self, with_log=True):
        """Initialize the dummy distribution.

        :param with_log: whether to attach a ``log_pdf`` attribute.
        """
        self.label = "dummy"
        self.names = ["p0", "p1"]
        self.labels = ["p0", "p1"]
        self.lims = {"p0": [-1.0, 1.0], "p1": [-1.0, 1.0]}
        self.with_log = with_log
        if with_log:
            self.log_pdf = self._log_pdf

    def sim(self, n):
        """Draw samples from the dummy distribution.

        :param n: number of samples to draw.
        :returns: samples with shape ``(n, 2)``.
        """
        rng = np.random.default_rng(42)
        return rng.normal(scale=0.5, size=(n, 2))

    def pdf(self, x):
        """Evaluate the probability density.

        :param x: sample coordinates.
        :returns: density values for ``x``.
        """
        arr = np.asarray(x)
        return np.exp(-0.5 * np.sum(arr**2, axis=-1)) / (2 * np.pi)

    def _log_pdf(self, x):
        """Evaluate the log-probability density.

        :param x: sample coordinates.
        :returns: log-density values for ``x``.
        """
        arr = np.asarray(x)
        return -0.5 * np.sum(arr**2, axis=-1) - np.log(2 * np.pi)


#########################################################################################################
# Wrapper tests


class TestTfProbWrapper(unittest.TestCase):
    """TensorFlow Probability wrapper test suite."""
    def test_wrapper_uses_distribution(self):
        """Test wrapper uses distribution metadata."""
        dist = tfp.distributions.MultivariateNormalDiag(
            loc=[0.0, 0.0], scale_diag=[1.0, 2.0], name="mvdiag"
        )
        wrapper = af.tf_prob_wrapper(dist)
        coords = tf.constant([[0.5, -0.5]], dtype=tf.float32)
        logp = wrapper.log_pdf(coords)
        self.assertEqual(logp.shape, (1,))
        samples = wrapper.sim(3)
        self.assertEqual(samples.shape, (3, 2))
        self.assertEqual(wrapper.names, ["p0", "p1"])


#########################################################################################################
# Analytic flow tests


class TestAnalyticFlow(unittest.TestCase):
    """Analytic flow test suite."""
    def test_invalid_inputs_raise(self):
        """Test invalid input validation."""
        class MissingSim:
            """Missing Sim test suite."""
            def pdf(self, x):
                """Pdf."""
                return x

        class MissingPdf:
            """Missing Pdf test suite."""
            def sim(self, n):
                """Sim."""
                return np.zeros((n, 1))

        with self.assertRaises(ValueError):
            af.analytic_flow(MissingSim())
        with self.assertRaises(ValueError):
            af.analytic_flow(MissingPdf())
        self.assertTrue(np.allclose(MissingSim().pdf(np.array([1.0])), np.array([1.0])))
        self.assertEqual(MissingPdf().sim(1).shape, (1, 1))

    def test_log_probability_and_derivatives(self):
        """Test log-probability values and derivatives."""
        flow = af.analytic_flow(DummyDistribution(with_log=True))
        test_point = tf.constant([[0.0, 0.0]], dtype=tf.float32)
        lp = flow.log_probability(test_point)
        self.assertTrue(np.all(np.isfinite(lp.numpy())))
        jac = flow.log_probability_jacobian(np.array([0.1, -0.1]))
        self.assertEqual(jac.shape, (2,))
        hess = flow.log_probability_hessian(np.array([0.2, -0.2]))
        self.assertEqual(hess.shape, (2, 2))
        samples = flow.sample(4)
        self.assertEqual(samples.shape, (4, 2))
        self.assertIsNone(flow.reset_tensorflow_caches())

    def test_branch_without_log_pdf(self):
        """Test branch that falls back to pdf without log_pdf."""
        flow = af.analytic_flow(DummyDistribution(with_log=False))
        coords = np.array([[0.1, 0.2], [0.0, 0.0]])
        lp = flow.log_probability(coords).numpy()
        manual = np.log(flow._dist.pdf(coords))
        np.testing.assert_allclose(lp, manual)
        jac = flow.log_probability_jacobian(coords)
        self.assertEqual(jac.shape, coords.shape)

    def test_mcsamples_output(self):
        """Test MCSamples output shape and metadata."""
        flow = af.analytic_flow(DummyDistribution())
        mc = flow.MCSamples(6)
        self.assertEqual(mc.samples.shape[0], 6)
        self.assertEqual(len(mc.loglikes), 6)
        running = mc.getParamNames().getRunningNames()
        self.assertEqual(set(running), {"p0", "p1"})

#########################################################################################################
# Additional helper distributions


class DummyNumpyDist:
    """Dummy Numpy Dist test suite."""
    def __init__(self):
        """Init."""
        self.label = "dummy"
        self.names = ["x"]
        self.labels = ["x"]
        self.lims = {"x": (0.0, 1.0)}

    def sim(self, n):
        """Simulate samples.

        :param n: number of samples.
        :returns: samples with shape ``(n, 1)``.
        """
        return np.ones((n, 1))

    def pdf(self, x):
        """Evaluate the probability density.

        :param x: sample coordinates.
        :returns: density values for ``x``.
        """
        return np.ones_like(x)


class DummyNumpyLogPDF(DummyNumpyDist):
    """Dummy Numpy Log P D F test suite."""
    def log_pdf(self, x):
        """Evaluate the log density.

        :param x: sample coordinates.
        :returns: log-density values for ``x``.
        """
        return np.zeros_like(x)

#########################################################################################################
# Additional analytic flow tests


class TestAnalyticFlowAdditional(unittest.TestCase):
    """Analytic flow additional test suite."""
    def test_tf_prob_wrapper(self):
        """Test TensorFlow Probability wrapper precisions."""
        dist = tfd.MultivariateNormalDiag(loc=[0.0, 0.0], scale_diag=[1.0, 2.0])
        wrapper = af.tf_prob_wrapper(dist, prec=tf.float32)
        samp = wrapper.sim(2)
        self.assertEqual(samp.shape[-1], 2)
        lp = wrapper.log_pdf(tf.constant([[0.0, 0.0]], dtype=tf.float32))
        self.assertEqual(lp.shape, (1,))

    def test_init_requires_pdf(self):
        """Test initialization requires pdf attribute."""
        class BadDist:
            """Bad Dist test suite."""
            def sim(self, n):
                """Simulate samples.

                :param n: number of samples.
                :returns: samples with shape ``(n, 1)``.
                """
                return np.zeros((n, 1))
        self.assertEqual(BadDist().sim(2).shape, (2, 1))
        with self.assertRaises(ValueError):
            af.analytic_flow(BadDist())

    def test_init_without_log_pdf(self):
        """Test initialization without log_pdf attribute."""
        dist = DummyNumpyDist()
        flow = af.analytic_flow(dist)
        self.assertFalse(flow.has_log_pdf)

    def test_log_probability_branches(self):
        """Test log_probability branches with and without log_pdf."""
        flow_with_log = af.analytic_flow(DummyNumpyLogPDF())
        flow_without_log = af.analytic_flow(DummyNumpyDist())
        coords = np.array([[0.0], [0.5]])
        res_with = flow_with_log.log_probability(coords)
        res_without = flow_without_log.log_probability(coords)
        self.assertTrue(np.allclose(res_with.numpy(), 0.0))
        self.assertTrue(np.allclose(res_without.numpy(), 0.0))

    def test_log_probability_tensor_input(self):
        """Test tensor input handling in log_probability."""
        flow = af.analytic_flow(DummyNumpyLogPDF())
        coords = tf.constant([[0.1], [0.2]], dtype=tf.float32)
        res = flow.log_probability(coords)
        self.assertEqual(res.shape[0], 2)

    def test_jacobian_and_hessian(self):
        """Test Jacobian and Hessian calculations."""
        flow = af.analytic_flow(DummyNumpyLogPDF())
        coords = np.array([[0.1], [0.2]])
        jac = flow.log_probability_jacobian(coords)
        self.assertEqual(jac.shape[0], 2)
        hess = flow.log_probability_hessian(coords)
        self.assertEqual(hess.shape[0], 2)
        jac_scalar = flow.log_probability_jacobian(coords[0])
        hess_scalar = flow.log_probability_hessian(coords[0])
        self.assertEqual(jac_scalar.shape[0], 1)
        self.assertEqual(hess_scalar.shape[0], 1)

    def test_mcsamples_generation(self):
        """Test MCSamples generation with log-likelihoods."""
        flow = af.analytic_flow(DummyNumpyLogPDF())
        mc = flow.MCSamples(size=3, logLikes=True)
        self.assertEqual(mc.samples.shape[0], 3)
        self.assertIsNotNone(mc.loglikes)

    def test_reset_placeholder(self):
        """Test reset_tensorflow_caches placeholder."""
        flow = af.analytic_flow(DummyNumpyLogPDF())
        self.assertIsNone(flow.reset_tensorflow_caches())

    def test_custom_metadata_overrides_defaults(self):
        """Test metadata overrides for analytic_flow inputs."""
        class CustomDist(DummyNumpyLogPDF):
            """Custom Dist test suite."""
            def __init__(self):
                """Init."""
                super().__init__()
                self.label = "base"
                self.names = ["a"]
                self.labels = ["a"]
                self.lims = {"a": (-1.0, 1.0)}

        dist = CustomDist()
        flow = af.analytic_flow(
            dist,
            name_tag="override",
            param_names=["x"],
            param_labels=["x_label"],
            lims={"x": (0.0, 2.0)},
        )
        self.assertEqual(flow.name_tag, "override")
        self.assertEqual(flow.param_names, ["x"])
        self.assertEqual(flow.param_labels, ["x_label"])
        self.assertEqual(flow.parameter_ranges, {"x": (0.0, 2.0)})

    def test_mcsamples_without_loglikes(self):
        """Test MCSamples path without loglikes."""
        flow = af.analytic_flow(DummyNumpyLogPDF())

        class DummyMCSamples:
            """Dummy MCSamples test suite."""
            def __init__(self, samples, loglikes, **kwargs):
                """Init."""
                self.samples = samples
                self.loglikes = loglikes

        dummy = DummyMCSamples(np.zeros((1, 1)), None)
        self.assertEqual(dummy.samples.shape, (1, 1))

        with patch("tensiometer.synthetic_probability.analytic_flow.MCSamples", DummyMCSamples):
            with self.assertRaises(AttributeError):
                flow.MCSamples(size=2, logLikes=False)

    def test_jacobian_tensor_without_log_pdf(self):
        """Test tensor Jacobian without log_pdf."""
        flow = af.analytic_flow(DummyNumpyDist())
        coords = tf.constant([0.5], dtype=tf.float32)
        jac = flow.log_probability_jacobian(coords)
        self.assertEqual(jac.shape, (1,))

    def test_hessian_tensor_input(self):
        """Test tensor Hessian input handling."""
        flow = af.analytic_flow(DummyNumpyLogPDF())
        coords = tf.constant([[0.0]], dtype=tf.float32)
        hess = flow.log_probability_hessian(coords)
        self.assertEqual(hess.shape[0], 1)

    def test_tensor_jacobian_and_hessian_branches(self):
        """Test tensor Jacobian and Hessian branches."""
        flow = af.analytic_flow(DummyNumpyLogPDF())
        tensor_coords = tf.constant([[0.0]], dtype=tf.float32)
        jac = flow.log_probability_jacobian(tensor_coords)
        hess = flow.log_probability_hessian(tensor_coords)
        self.assertEqual(jac.shape[0], 1)
        self.assertEqual(hess.shape[0], 1)

    def test_scalar_jacobian_without_log_pdf(self):
        """Test scalar Jacobian without log_pdf."""
        flow = af.analytic_flow(DummyNumpyDist())
        jac = flow.log_probability_jacobian(np.array([0.5]))
        self.assertEqual(jac.shape, (1,))


#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""
File with tools to interface Cosmosis chains with GetDist.
"""

###############################################################################
# initial imports:

import functools
import os
import re
import types

import numpy as np
from scipy.special import logsumexp

from getdist.chains import loadNumpyTxt
from getdist.mcsamples import MCSamples
from getdist.paramnames import ParamInfo
from getdist.types import BestFit

############################################################################### 


def MCSamplesFromCosmosis(chain_root, chain_min_root=None,
                          param_name_dict=None, param_label_dict=None,
                          name_tag=None, settings=None):
    """
    Function to import Cosmosis chains in GetDist.

    :param chain_root: the name and path to the chain or the path to the
        folder that contains it.
    :param chain_min_root: (optional) name of the file containing the
        explicit best fit.
    :param param_name_dict: (optional) a dictionary with the mapping between
        cosmosis names and reasonable parameter names.
    :param param_label_dict: (optional) dictionary with the mapping between
        parameter names and parameter labels, since Cosmosis does not save
        the labels in the chain.
    :param name_tag: (optional) a string with the name tag for the chain.
    :param settings: (optional) dictionary of analysis settings to override
        getdist defaults
    :return: The :class:`~getdist.mcsamples.MCSamples` instance
    """
    # decide if the user passed a folder or a chain:
    if os.path.isfile(chain_root+'.txt'):
        chain_file = chain_root+'.txt'
    elif os.path.isdir(chain_root):
        # look for the chain file:
        temp = list(filter(lambda x: 'chain.txt' in x, os.listdir(chain_root)))
        if len(temp) == 0:
            raise ValueError('No chain file found in folder', chain_root)
        chain_file = chain_root+'/'+temp[0]
    else:
        raise ValueError('Input chain root is not a folder nor a file.')
    # get all the commented lines in the chain file:
    info = get_cosmosis_info(chain_file)
    # get the parameter names:
    param_names = get_param_names(info)
    # get the parameter labels from the user provided dictionary:
    param_labels = get_param_labels(info, param_names, param_label_dict)
    # get the sampler:
    sampler_type, sampler = get_sampler_type(info)
    # get the name tag:
    if name_tag is None:
        name_tag = get_name_tag(info)
    # get the samples weights and likelihood:
    chain = loadNumpyTxt(chain_file, skiprows=0)
    # parse the chain depending on the sampler that produced it:
    if sampler_type == 'nested':
        # get number of samples to use:
        nsamples = int(list(filter(lambda x: 'nsample=' in x, info))
                       [0].replace(' ', '').split('=')[1])
        # get the chain:
        chain = chain[-nsamples:]
        # decide whether we have weights or log weights:
        if 'weight' in param_names:
            _w = 'weight'
        elif 'log_weight' in param_names:
            _w = 'log_weight'
        else:
            raise ValueError('Nested sampler chains should have weights')
        # get all quantities:
        indexes = [i for i in range(len(param_names))
                   if i != param_names.index(_w)
                   and i != param_names.index('post')]
        samples = chain[:, indexes]
        weights = chain[:, param_names.index(_w)]
        if _w == 'log_weight':
            weights = weights - logsumexp(weights)
            weights = np.exp(weights)
        loglike = chain[:, param_names.index('post')]
        # delete the weights and likelihood from names:
        if param_labels is not None:
            param_labels.pop(param_labels.index(_w))
            param_labels.pop(param_labels.index('post'))
        param_names.pop(param_names.index(_w))
        param_names.pop(param_names.index('post'))
    elif sampler_type == 'mcmc':
        # get all quantities:
        indexes = [i for i in range(len(param_names))
                   if i != param_names.index('post')]
        samples = chain[:, indexes]
        loglike = chain[:, param_names.index('post')]
        # process weights:
        if sampler == 'importance':
            # get the old weights or the old_log_weights:
            if 'old_weight' in param_names:
                old_weights = chain[:, param_names.index('old_weight')]
            elif 'old_log_weight' in param_names:
                old_weights = np.exp(chain[:, param_names.index('old_log_weight')])
            # create filter:
            nonzeromask = old_weights!=0
            # get the IS weights:
            isweights = chain[:, param_names.index('log_weight')]
            # exp the filtered IS weights:
            wexp =np.exp(isweights[nonzeromask])
            # create the new weights:
            weights = np.zeros(old_weights.size)
            weights[nonzeromask] = old_weights[nonzeromask]*wexp
        else:
            # check for existing weights:
            if 'weight' in param_names or 'log_weight' in param_names:
                raise ValueError('MCMC chains should not have weights')
            # Cosmosis does not weight samples:
            samples, idx, weights = np.unique(samples, return_index=True, return_counts=True, axis=0)
            loglike = loglike[idx]
        # delete the weights and likelihood from names:
        if param_labels is not None:
            param_labels.pop(param_names.index('post'))
        param_names.pop(param_names.index('post'))
    elif sampler_type == 'uncorrelated':
        # get all quantities:
        indexes = [i for i in range(len(param_names))
                   if i != param_names.index('post')]
        samples = chain[:, indexes]
        loglike = chain[:, param_names.index('post')]
        weights = None
        # delete the weights and likelihood from names:
        if param_labels is not None:
            param_labels.pop(param_names.index('post'))
        param_names.pop(param_names.index('post'))
    else:
        raise ValueError('Unknown sampler')
    # get the ranges:
    ranges = get_ranges(info, param_names)
    # transform param names:
    if param_name_dict is not None:
        for i, name in enumerate(param_names):
            if name in param_name_dict.keys():
                param_names[i] = param_name_dict[name]
                if name in ranges.keys():
                    ranges[param_name_dict[name]] = ranges.pop(name)
    # initialize the samples:
    mc_samples = MCSamples(samples=samples, weights=weights,
                           loglikes=-loglike,
                           sampler=sampler_type, names=param_names,
                           labels=param_labels, ranges=ranges,
                           ignore_rows=0, name_tag=name_tag,
                           settings=settings)
    # set running parameters:
    for name in mc_samples.getParamNames().parsWithNames(
            mc_samples.getParamNames().list()):
        if name.name in ranges.keys():
            name.isDerived = False
        else:
            name.isDerived = True
    # polish the samples removing nans:
    mc_samples = polish_samples(mc_samples)
    # get the best fit:
    if chain_min_root is not None:
        # since getdist does not cache the best fit we have to override the
        # method in this brute way:
        funcType = types.MethodType
        mc_samples.getBestFit = funcType(functools.partial(get_maximum_likelihood,
                                                           chain_min_root=chain_min_root,
                                                           param_name_dict=param_name_dict,
                                                           param_label_dict=param_label_dict),
                                                           mc_samples)
    # update statistics:
    mc_samples.updateBaseStatistics()
    #
    return mc_samples

###############################################################################


def get_cosmosis_info(file):
    """
    Parse a file to get all the information about a Cosmosis run.

    :param file: path and name of the file to parse.
    :return: a list of strings with the cosmosis parameters for the run.
    """
    with open(file, 'r') as _file:
        info = []
        for ln in _file:
            if ln.startswith('#'):
                info.append(ln[1:])
    # polish the info strings:
    info = list(filter(None, [line.replace('#', '').strip() for line in info]))
    #
    return info

###############################################################################


def get_param_names(info):
    """
    Get the parameter names for a Cosmosis run.

    :param info: a list of strings with the cosmosis parameters for the run.
    :return: a list of strings with the parameter names.
    """
    return list(filter(None, re.split('\t| ', info[0])))

###############################################################################


def get_param_labels(info, param_names, param_label_dict):
    """
    Get the labels for the parameter names of a Cosmosis run.

    :param info: a list of strings with the cosmosis parameters for the run.
    :param param_names: a list of strings with the parameter names.
    :param param_label_dict: a dictionary with the mapping between names and
        labels.
    :return: a list of strings with the parameter labels.
    """
    if param_label_dict is not None:
        param_labels = []
        for name in param_names:
            if name in param_label_dict.keys():
                param_labels.append(param_label_dict[name])
            else:
                param_labels.append(name)
    else:
        param_labels = None
    #
    return param_labels

###############################################################################


def get_sampler_type(info):
    """
    Get the sampler type for a chain given the a list of strings containing the
    cosmosis run parameter informations.
    To process the sampler type the function defines internally a dictionary
    with the mapping from sampler name to sampler type.

    :param info: a list of strings with the cosmosis parameters for the run.
    :return: a string with the sampler type if any, otherwise returns none.
    """
    # define the dictionary with the mapping:
    sampler_dict = {
                    'polychord': 'nested',
                    'multinest': 'nested',
                    'nautilus': 'nested',
                    'apriori': 'uncorrelated',
                    'emcee': 'mcmc',
                    'importance': 'mcmc',
                    'pmaxlike': 'max_like',
                    'maxlike': 'max_like',
                    'max_like': 'max_like'
                    }
    # find the sampler in the parameters:
    temp = list(filter(lambda x: 'sampler' in x, info))
    if len(temp) > 0:
        sampler = temp[0].replace(' ', '').split('=')[1].lower()
        if sampler in sampler_dict.keys():
            sampler_type = sampler_dict[sampler]
        else:
            raise ValueError('Unknown input sampler', sampler)
    else:
        sampler_type = None
    #
    return sampler_type, sampler

###############################################################################


def get_name_tag(info):
    """
    Get the name tag for a chain given the a list of strings containing the
    cosmosis run parameter informations.

    :param info: a list of strings with the cosmosis parameters for the run.
    :return: a string with the name tag if any, otherwise returns none.
    """
    temp = list(filter(lambda x: 'run_name' in x, info))
    if len(temp) > 0:
        name_tag = temp[0].replace(' ', '').split('=')[1].lower()
        name_tag = name_tag.rstrip().lstrip()
    else:
        name_tag = None
    #
    return name_tag

###############################################################################


def get_ranges(info, param_names):
    """
    Get the ranges for the parameters from the info file.

    :param info: a list of strings with the cosmosis parameters for the run.
    :param param_names: a list with the parameter names.
    :return: a dictionary with the parameter ranges.
    """
    ranges = {}
    for name in param_names:
        try:
            # look for the section:
            section = '['+name.split('--')[0]+']'
            pname = name.split('--')[1]
            section_index = info.index(section)
            temp = list(filter(lambda x:
                               pname in x[1] and x[0] > section_index,
                               enumerate(info)))[0][1]
            _min = float(list(filter(None, temp.split()))[2])
            _max = float(list(filter(None, temp.split()))[4])
            ranges[name] = [_min, _max]
        except:
            pass
    #
    return ranges

###############################################################################


def polish_samples(chain):
    """
    Remove fixed parameters and samples with some parameter that is Nan
    from the input chain.

    :param chain: :class:`~getdist.mcsamples.MCSamples` the input chain.
    :return: :class:`~getdist.mcsamples.MCSamples` the polished chain.
    """
    # delete fixed parameters:
    chain.deleteFixedParams()
    # we need to delete nan columns:
    fixed = []
    values = []
    for i in range(chain.samples.shape[1]):
        if np.all(np.isnan(chain.samples[:, i])):
            fixed.append(i)
            values.append(chain.samples[0, i])
    chain.changeSamples(np.delete(chain.samples, fixed, 1))
    if hasattr(chain, 'ranges'):
        for ix, value in zip(fixed, values):
            chain.ranges.setFixed(chain.paramNames.names[ix].name,
                                  value)
    chain.paramNames.deleteIndices(fixed)
    chain._getParamIndices()
    # now we need to delete nan samples:
    nan_filter = np.all(np.logical_not(np.isnan(chain.samples)), axis=1)
    chain.filter(where=nan_filter)
    #
    return chain

###############################################################################


def get_maximum_likelihood(dummy, max_posterior, chain_min_root,
                           param_name_dict, param_label_dict):
    """
    Import the maximum likelihood file for a Cosmosis run, if present.

    :param dummy: dummy argument for interfacing, not used in practice
    :param chain_min_root: name of the minimum file or the folder that contains
        it.
    :param param_name_dict: a dictionary with the mapping between
        cosmosis names and reasonable names.
    :param param_label_dict: dictionary with the mapping between the parameter
        names and the labels.
    :return: :class:`~getdist.types.BestFit` the best fit object.
    """
    # decide if the user passed a folder or a chain:
    if os.path.isfile(chain_min_root+'.txt'):
        minimum_file = chain_min_root+'.txt'
    elif os.path.isdir(chain_min_root):
        # look for the chain file:
        temp = list(filter(lambda x: 'chain_pmaxlike.txt' in x,
                    os.listdir(chain_min_root)))
        if len(temp) == 0:
            raise ValueError('No minimum file found in folder', chain_min_root)
        minimum_file = chain_min_root+'/'+temp[0]
    else:
        raise ValueError('Input chain root is not a folder nor a file.')
    # get the info:
    info = get_cosmosis_info(minimum_file)
    # check the sampler:
    sampler_type, sampler = get_sampler_type(info)
    if sampler_type != 'max_like':
        raise ValueError('Minimum file appears to have a strange sampler:',
                         sampler_type)
    # get the parameter names:
    param_names = get_param_names(info)
    # get the parameter labels from the user provided dictionary:
    param_labels = get_param_labels(info, param_names, param_label_dict)
    # best fit:
    best_fit = BestFit()
    # set parameter names:
    if param_name_dict is not None:
        best_fit.names = []
        for name in param_names:
            if name in param_name_dict.keys():
                best_fit.names.append(ParamInfo(param_name_dict[name]))
            else:
                best_fit.names.append(ParamInfo(name))
    else:
        best_fit.names = [ParamInfo(name) for name in param_names]
    if param_labels is not None:
        for name, label in zip(best_fit.names, param_labels):
            name.label = label
    # get the numbers:
    best_fit_params = loadNumpyTxt(minimum_file, skiprows=0)[0]
    # look for weight:
    if 'weigth' in param_names:
        best_fit.weight = best_fit_params[param_names.index('weigth')]
    else:
        best_fit.weight = 1.0
    # look for log like:
    if 'post' in param_names:
        best_fit.logLike = -best_fit_params[param_names.index('post')]
    else:
        raise ValueError('Best fit file does not contain the best fit like')
    # correct log likes in getdist format:
    if 'post' in param_names:
        best_fit_params[param_names.index('post')] = \
            -best_fit_params[param_names.index('post')]
    if 'prior' in param_names:
        best_fit_params[param_names.index('prior')] = \
            -best_fit_params[param_names.index('prior')]
    if 'like' in param_names:
        best_fit_params[param_names.index('like')] = \
            -best_fit_params[param_names.index('like')]
    # get the data chi squareds:
    best_fit.chiSquareds = []
    # get the ranges to get the running parameters:
    ranges = get_ranges(info, param_names)
    # now initialize all parameters:
    for ind, param in enumerate(best_fit.names):
        param.isDerived = param.name not in ranges.keys()
        param.number = ind
        param.best_fit = best_fit_params[ind]
    #
    return best_fit



"""Tests for tensor eigenvalue utilities."""

#########################################################################################################
# Imports

import unittest
from unittest.mock import patch

import numpy as np

import tensiometer.utilities.tensor_eigenvalues as te

#########################################################################################################
# Tensor Eigenvalue Utilities tests


class TestTensorEigenvalueUtilities(unittest.TestCase):

    """Tensor eigenvalue utility test suite."""
    def setUp(self):
        """Set up test fixtures."""
        pass

    def test_random_symm_tensor(self):
        """Test Random symm tensor."""
        assert te.random_symm_tensor(d=2, m=2).shape == (2, 2)
        assert te.random_symm_tensor(d=2, m=4).shape == (2, 2, 2, 2)
        assert te.random_symm_tensor(d=4, m=2).shape == (4, 4)
        assert te.random_symm_tensor(d=4, m=4).shape == (4, 4, 4, 4)
        temp = te.random_symm_tensor(d=8, m=2)
        assert np.allclose(temp, temp.T)
        temp = te.random_symm_tensor(d=8, m=4)
        assert np.allclose(temp, temp.T)

    def test_random_symm_positive_tensor(self):
        """Test Random symm positive tensor."""
        assert te.random_symm_positive_tensor(d=2, m=2).shape == (2, 2)
        assert te.random_symm_positive_tensor(d=2, m=4).shape == (2, 2, 2, 2)
        assert te.random_symm_positive_tensor(d=4, m=2).shape == (4, 4)
        assert te.random_symm_positive_tensor(d=4, m=4).shape == (4, 4, 4, 4)
        temp = te.random_symm_positive_tensor(d=8, m=2)
        assert np.allclose(temp, temp.T)
        temp = te.random_symm_positive_tensor(d=8, m=4)
        assert np.allclose(temp, temp.T)

    def test_identity_tensor(self):
        """Test Identity tensor."""
        assert te.identity_tensor(d=2, m=2).shape == (2, 2)

#########################################################################################################
# Random Symm Positive Tensor tests

class TestRandomSymmPositiveTensor(unittest.TestCase):

    """Test Random Symm Positive Tensor test suite."""
    def test_random_symm_positive_tensor(self):
        """Test Random symm positive tensor."""
        tensor = te.random_symm_positive_tensor(d=2, m=2)
        self.assertEqual(tensor.shape, (2, 2))
        
        tensor = te.random_symm_positive_tensor(d=2, m=4)
        self.assertEqual(tensor.shape, (2, 2, 2, 2))
        
        tensor = te.random_symm_positive_tensor(d=4, m=2)
        self.assertEqual(tensor.shape, (4, 4))
        
        tensor = te.random_symm_positive_tensor(d=4, m=4)
        self.assertEqual(tensor.shape, (4, 4, 4, 4))
        
        temp = te.random_symm_positive_tensor(d=8, m=2)
        self.assertTrue(np.allclose(temp, temp.T))
        
        temp = te.random_symm_positive_tensor(d=8, m=4)
        self.assertTrue(np.allclose(temp, temp.T))

#########################################################################################################
# Identity Tensor tests

class TestIdentityTensor(unittest.TestCase):

    """Test Identity Tensor test suite."""
    def test_identity_tensor(self):
        """Test Identity tensor."""
        tensor = te.identity_tensor(d=2, m=2)
        self.assertEqual(tensor.shape, (2, 2))

        tensor = te.identity_tensor(d=3, m=3)
        self.assertEqual(tensor.shape, (3, 3, 3))

        tensor = te.identity_tensor(d=4, m=2)
        self.assertEqual(tensor.shape, (4, 4))

        tensor = te.identity_tensor(d=5, m=4)
        self.assertEqual(tensor.shape, (5, 5, 5, 5))

        tensor = te.identity_tensor(d=6, m=3)
        self.assertEqual(tensor.shape, (6, 6, 6))

        tensor = te.identity_tensor(d=7, m=2)
        self.assertEqual(tensor.shape, (7, 7))

        tensor = te.identity_tensor(d=8, m=4)
        self.assertEqual(tensor.shape, (8, 8, 8, 8))

        tensor = te.identity_tensor(d=9, m=3)
        self.assertEqual(tensor.shape, (9, 9, 9))

        tensor = te.identity_tensor(d=10, m=2)
        self.assertEqual(tensor.shape, (10, 10))

#########################################################################################################
# Number Eigenvalues tests

class TestNumberEigenvalues(unittest.TestCase):

    """Test Number Eigenvalues test suite."""
    def test_number_eigenvalues(self):
        """Test Number eigenvalues."""
        self.assertEqual(te.number_eigenvalues(d=2, m=2), 2)
        self.assertEqual(te.number_eigenvalues(d=2, m=4), 6)
        self.assertEqual(te.number_eigenvalues(d=4, m=2), 4)
        self.assertEqual(te.number_eigenvalues(d=4, m=4), 108)
        self.assertEqual(te.number_eigenvalues(d=8, m=2), 8)
        self.assertEqual(te.number_eigenvalues(d=8, m=4), 17496)

#########################################################################################################
# Tensor Deflation tests

class TestTensorDeflation(unittest.TestCase):

    """Test Tensor Deflation test suite."""
    def setUp(self):
        """ Set up test data. """
        self.A_2D = np.array([[1, 2], [3, 4]])  # A simple 2x2 matrix
        self.A_3D = np.random.rand(3, 3, 3)    # A random 3x3x3 tensor
        self.x_2D = np.array([1, 1])           # A vector of ones for 2D
        self.x_3D = np.array([1, 0, -1])       # A sample vector for 3D
        self.l = 2                             # Scalar to deflate by

    def test_tensor_deflation_2D(self):
        """ Test deflation on a 2D matrix with a simple vector. """
        result = te.tensor_deflation(self.A_2D, self.l, self.x_2D)
        
        # Manually calculating expected deflation
        outer_product = np.outer(self.x_2D, self.x_2D)
        expected = self.A_2D - self.l * outer_product
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_deflation_3D(self):
        """ Test deflation on a 3D tensor with a vector. """
        result = te.tensor_deflation(self.A_3D, self.l, self.x_3D)
        
        # Construct the outer product manually for 3D
        outer_product = np.multiply.outer(self.x_3D, np.outer(self.x_3D, self.x_3D))
        expected = self.A_3D - self.l * outer_product
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_deflation_zero_scalar(self):
        """ Test deflation when scalar is zero (should return the original tensor). """
        result = te.tensor_deflation(self.A_2D, 0, self.x_2D)
        
        # If scalar is zero, no deflation should occur
        expected = self.A_2D
        
        # Asserting equality
        np.testing.assert_array_equal(result, expected)

    def test_tensor_deflation_zero_vector(self):
        """ Test deflation with a zero vector (should return the original tensor). """
        zero_vector = np.zeros_like(self.x_2D)
        result = te.tensor_deflation(self.A_2D, self.l, zero_vector)
        
        # With a zero vector, no deflation should occur
        expected = self.A_2D
        
        # Asserting equality
        np.testing.assert_array_equal(result, expected)


#########################################################################################################
# Tensor Contraction tests

class TestTensorContraction(unittest.TestCase):

    """Test Tensor Contraction test suite."""
    def setUp(self):
        """ Set up test data. """
        # A simple 2x2 matrix (rank 2 tensor)
        self.A_2D = np.array([[1, 2], [3, 4]])
        # A random 3x3x3 tensor (rank 3 tensor)
        self.A_3D = np.random.rand(3, 3, 3)
        # Sample vectors for contraction
        self.x_2D = np.array([1, 1])      # For contracting a 2D tensor
        self.x_3D = np.array([1, 0, -1])  # For contracting a 3D tensor

    def test_tensor_contraction_2D_n1(self):
        """ Test contraction of a 2D tensor with a vector, contracted once (n=1). """
        result = te.tensor_contraction_brute_1(self.A_2D, self.x_2D, n=1)
        
        # Expected result: dot product of A and x_2D
        expected = np.dot(self.A_2D, self.x_2D)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_2D_n2(self):
        """ Test full contraction of a 2D tensor (contracted twice, n=2). """
        result = te.tensor_contraction_brute_1(self.A_2D, self.x_2D, n=2)
        
        # Expected result: dot product of (A * x_2D) and x_2D
        intermediate = np.dot(self.A_2D, self.x_2D)
        expected = np.dot(intermediate, self.x_2D)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_3D_n1(self):
        """ Test contraction of a 3D tensor with a vector, contracted once (n=1). """
        result = te.tensor_contraction_brute_1(self.A_3D, self.x_3D, n=1)
        
        # Expected result: contraction along the first axis (dot product with x_3D)
        expected = np.tensordot(self.A_3D, self.x_3D, axes=1)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_3D_n2(self):
        """ Test contraction of a 3D tensor with a vector, contracted twice (n=2). """
        result = te.tensor_contraction_brute_1(self.A_3D, self.x_3D, n=2)
        
        # Expected result: contract the tensor twice
        intermediate = np.tensordot(self.A_3D, self.x_3D, axes=1)
        expected = np.tensordot(intermediate, self.x_3D, axes=1)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_3D_n3(self):
        """ Test full contraction of a 3D tensor (contracted three times, n=3). """
        result = te.tensor_contraction_brute_1(self.A_3D, self.x_3D, n=3)
        
        # Expected result: fully contracted into a scalar
        intermediate_1 = np.tensordot(self.A_3D, self.x_3D, axes=1)
        intermediate_2 = np.tensordot(intermediate_1, self.x_3D, axes=1)
        expected = np.dot(intermediate_2, self.x_3D)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_zero_contracts(self):
        """ Test contraction with n=0 (should return the original tensor). """
        result = te.tensor_contraction_brute_1(self.A_2D, self.x_2D, n=0)
        
        # No contraction, so the result should be the original tensor
        expected = self.A_2D
        
        # Asserting equality
        np.testing.assert_array_equal(result, expected)

#########################################################################################################
# Tensor Contraction Brute2 tests

class TestTensorContractionBrute2(unittest.TestCase):

    """Test Tensor Contraction Brute2 test suite."""
    def setUp(self):
        """ Set up test data. """
        # A simple 2x2 matrix (rank 2 tensor)
        self.A_2D = np.array([[1, 2], [3, 4]])
        # A random 3x3x3 tensor (rank 3 tensor)
        self.A_3D = np.random.rand(3, 3, 3)
        # Sample vectors for contraction
        self.x_2D = np.array([1, 1])      # For contracting a 2D tensor
        self.x_3D = np.array([1, 0, -1])  # For contracting a 3D tensor

    def test_tensor_contraction_2D_n1(self):
        """ Test contraction of a 2D tensor with a vector, contracted once (n=1). """
        result = te.tensor_contraction_brute_2(self.A_2D, self.x_2D, n=1)
        
        # Expected result: dot product of A and x_2D
        expected = np.dot(self.A_2D, self.x_2D)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_2D_n2(self):
        """ Test full contraction of a 2D tensor (contracted twice, n=2). """
        result = te.tensor_contraction_brute_2(self.A_2D, self.x_2D, n=2)
        
        # Expected result: dot product of (A * x_2D) and x_2D
        intermediate = np.dot(self.A_2D, self.x_2D)
        expected = np.dot(intermediate, self.x_2D)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_3D_n1(self):
        """ Test contraction of a 3D tensor with a vector, contracted once (n=1). """
        result = te.tensor_contraction_brute_2(self.A_3D, self.x_3D, n=1)
        
        # Expected result: contraction along the first axis (dot product with x_3D)
        expected = np.tensordot(self.A_3D, self.x_3D, axes=1)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_3D_n2(self):
        """ Test contraction of a 3D tensor with a vector, contracted twice (n=2). """
        result = te.tensor_contraction_brute_2(self.A_3D, self.x_3D, n=2)
        
        # Expected result: contract the tensor twice
        intermediate = np.tensordot(self.A_3D, self.x_3D, axes=1)
        expected = np.tensordot(intermediate, self.x_3D, axes=1)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_3D_n3(self):
        """ Test full contraction of a 3D tensor (contracted three times, n=3). """
        result = te.tensor_contraction_brute_2(self.A_3D, self.x_3D, n=3)
        
        # Expected result: fully contracted into a scalar
        intermediate_1 = np.tensordot(self.A_3D, self.x_3D, axes=1)
        intermediate_2 = np.tensordot(intermediate_1, self.x_3D, axes=1)
        expected = np.dot(intermediate_2, self.x_3D)
        
        # Asserting equality
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_tensor_contraction_zero_contracts(self):
        """ Test contraction with n=0 (should return the original tensor). """
        result = te.tensor_contraction_brute_2(self.A_2D, self.x_2D, n=0)
        
        # No contraction, so the result should be the original tensor
        expected = self.A_2D
        
        # Asserting equality
        np.testing.assert_array_equal(result, expected)

#########################################################################################################
# Sphere Optimization tests

class TestSphereOptimization(unittest.TestCase):

    """Test Sphere Optimization test suite."""
    def setUp(self):
        """ Set up test data. """
        # A simple 3D unit vector on the sphere
        self.x = np.array([1.0, 0.0, 0.0])
        # Random 3D vector to simulate a Euclidean gradient
        self.egrad = np.array([0.5, 0.1, -0.2])
        # Random 3x3 Hessian matrix
        self.ehess = np.array([[0.2, 0.1, 0.05],
                               [0.1, 0.3, -0.1],
                               [0.05, -0.1, 0.4]])
        # Tangent vector for the sphere
        self.u = np.array([0.0, 1.0, 0.0])

    def test_eu_to_sphere_grad(self):
        """ Test Euclidean gradient conversion to spherical gradient. """
        result = te.eu_to_sphere_grad(self.x, self.egrad)
        
        # Expected gradient projection: egrad - <x, egrad> * x
        dot_product = np.dot(self.x, self.egrad)
        expected = self.egrad - dot_product * self.x
        
        # Asserting the spherical gradient is computed correctly
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_eu_to_sphere_grad_orthogonality(self):
        """ Test if the spherical gradient is orthogonal to the vector x (tangent space). """
        result = te.eu_to_sphere_grad(self.x, self.egrad)
        
        # The result should be orthogonal to x, i.e., dot(result, x) == 0
        dot_product = np.dot(result, self.x)
        
        # Assert that the spherical gradient is orthogonal to the vector on the sphere
        self.assertAlmostEqual(dot_product, 0.0, places=5)

    def test_eu_to_sphere_grad_zero_gradient(self):
        """ Test if the spherical gradient is zero when egrad is zero. """
        zero_egrad = np.zeros_like(self.egrad)
        result = te.eu_to_sphere_grad(self.x, zero_egrad)
        
        # The gradient should also be zero
        expected = zero_egrad
        
        # Asserting the result is zero
        np.testing.assert_array_equal(result, expected)

    def test_eu_to_sphere_hess(self):
        """ Test the Euclidean Hessian conversion to spherical Hessian. """
        result = te.eu_to_sphere_hess(self.x, self.egrad, self.ehess, self.u)
        
        # Manually compute the expected spherical Hessian
        ehess_dot_u = np.dot(self.ehess, self.u)
        expected = ehess_dot_u - np.dot(self.x, ehess_dot_u) * self.x \
                   - np.dot(self.x, self.egrad) * self.u
        
        # Asserting the spherical Hessian is computed correctly
        np.testing.assert_array_almost_equal(result, expected, decimal=5)

    def test_eu_to_sphere_hess_orthogonality(self):
        """ Test if the spherical Hessian is orthogonal to the vector x (tangent space). """
        result = te.eu_to_sphere_hess(self.x, self.egrad, self.ehess, self.u)
        
        # The result should be orthogonal to x, i.e., dot(result, x) == 0
        dot_product = np.dot(result, self.x)
        
        # Assert that the spherical Hessian is orthogonal to the vector on the sphere
        self.assertAlmostEqual(dot_product, 0.0, places=5)
        
#########################################################################################################
# Tensor Eigenvalues Extra tests


class TestTensorEigenvaluesExtra(unittest.TestCase):
    """Tensor eigenvalues extra test suite."""
    def test_random_symmetry_helpers(self):
        """Test Random symmetry helpers."""
        tens = te.random_symm_tensor(3, 3, vmin=0.0, vmax=0.1)
        self.assertTrue(np.allclose(tens, np.transpose(tens, axes=(1, 0, 2))))
        pos_tens = te.random_symm_positive_tensor(2, 4, vmin=0.0, vmax=0.1)
        self.assertTrue(np.all(pos_tens >= 0))
        ident = te.identity_tensor(2, 3)
        self.assertEqual(ident.shape, (2, 2, 2))
        self.assertEqual(te.number_eigenvalues(2, 3), 4)

    def test_tensor_contractions_and_deflation(self):
        """Test Tensor contractions and deflation."""
        A = np.ones((2, 2, 2))
        x = np.array([1.0, 2.0])
        res1 = te.tensor_contraction_brute_1(A, x, n=2)
        res2 = te.tensor_contraction_brute_2(A, x, n=2)
        self.assertTrue(np.allclose(res1, res2))
        deflated = te.tensor_deflation(A, l=1.5, x=np.array([1.0, 0.0]))
        self.assertEqual(deflated.shape, A.shape)

    def test_sphere_transforms(self):
        """Test Sphere transforms."""
        x = np.array([1.0, 0.0])
        egrad = np.array([2.0, 3.0])
        ehess = np.array([[1.0, 0.0], [0.0, 1.0]])
        u = np.array([0.0, 1.0])
        grad = te.eu_to_sphere_grad(x, egrad)
        self.assertAlmostEqual(np.dot(grad, x), 0.0)
        hess = te.eu_to_sphere_hess(x, egrad, ehess, u)
        self.assertEqual(hess.shape, (2,))

    def test_rayleigh_helpers(self):
        """Test Rayleigh helpers."""
        A = np.eye(2)
        x = np.array([1.0, 0.0])
        self.assertAlmostEqual(te.tRq(x, A), 1.0)
        res = te.tRq_nder(x, A, 1)
        self.assertTrue(np.allclose(res, np.array([2.0, 0.0])))

    def test_brute_autograd_and_max(self):
        """Test Brute autograd and max."""
        A = np.eye(2)
        # monkeypatch solvers to avoid heavy optimization
        original_ps = te.pymanopt.solvers.ParticleSwarm
        original_tr = te.pymanopt.solvers.TrustRegions

        class DummySolver:
            """Dummy Solver test suite."""
            def __init__(self, *args, **kwargs):
                """Init."""
                pass

            def solve(self, problem):
                """Solve."""
                return np.array([1.0, 0.0])

        te.pymanopt.solvers.ParticleSwarm = DummySolver
        te.pymanopt.solvers.TrustRegions = DummySolver
        try:
            val, vec = te.max_tRq_brute(A, optimizer="ParticleSwarm")
            self.assertAlmostEqual(val, 1.0)
            self.assertTrue(np.allclose(vec, np.array([1.0, 0.0])))
            val2, vec2 = te.max_GtRq_brute(A, A, optimizer="TrustRegions")
            self.assertTrue(np.isfinite(val2))
            self.assertEqual(vec2.shape[0], 2)
        finally:
            te.pymanopt.solvers.ParticleSwarm = original_ps
            te.pymanopt.solvers.TrustRegions = original_tr

    def test_brute_2d(self):
        """Test Brute 2d."""
        A = np.eye(2)
        eig, eigv = te.tRq_brute_2D(A, num_points=10)
        self.assertEqual(eig.shape[0], eigv.shape[0])
        eig2, eigv2 = te.GtRq_brute_2D(A, A, num_points=10)
        self.assertEqual(eig2.shape[0], eigv2.shape[0])

    def test_power_method(self):
        """Test Power method."""
        A = np.eye(2)
        val, vec = te.max_tRq_power(A, maxiter=5, tol=1e-6, x0=np.array([1.0, 0.0]), history=False)
        self.assertAlmostEqual(val, 1.0)
        self.assertTrue(np.allclose(vec, np.array([1.0, 0.0])))


class TestTensorEigenvaluesBranch(unittest.TestCase):
    """Tensor eigenvalues branch test suite."""
    def test_max_tRq_brute_trustregions_branch(self):
        """Test Max t Rq brute trustregions branch."""
        class DummySolver:
            """Dummy Solver test suite."""
            def __init__(self, *args, **kwargs):
                """Init."""
                pass

            def solve(self, problem):
                """Solve."""
                return np.array([1.0, 0.0])

        with patch.object(te.pymanopt.solvers, "TrustRegions", DummySolver):
            val, vec = te.max_tRq_brute(np.eye(2), optimizer="TrustRegions")

        self.assertAlmostEqual(val, 1.0)
        self.assertTrue(np.allclose(vec, np.array([1.0, 0.0])))

    def test_max_tRq_power_warning_and_history(self):
        """Test Max t Rq power warning and history."""
        np.random.seed(0)
        A = np.array([[2.0, 0.0], [0.0, 1.0]])
        val, vec, history = te.max_tRq_power(A, maxiter=1, tol=0.0, x0=None, history=True)
        self.assertEqual(len(history), 1)
        self.assertEqual(vec.shape, (2,))
        self.assertTrue(np.isfinite(val))

    def test_max_tRq_shift_power_warning_and_history(self):
        """Test Max t Rq shift power warning and history."""
        A = np.array([[2.0, 0.0], [0.0, 1.0]])
        val, vec, history = te.max_tRq_shift_power(A, alpha=0.1, maxiter=1, tol=0.0, x0=None, history=True)
        self.assertEqual(len(history), 1)
        self.assertEqual(vec.shape, (2,))
        self.assertTrue(np.isfinite(val))

    def test_max_tRq_geap_warning_and_history(self):
        """Test Max t Rq geap warning and history."""
        A = np.eye(2)
        val, vec, history = te.max_tRq_geap(A, maxiter=1, tol=0.0, x0=None, history=True)
        self.assertEqual(history.shape[0], 1)
        self.assertEqual(vec.shape, (2,))
        self.assertTrue(np.isfinite(val))

    def test_max_tRq_shift_power_break_and_no_history(self):
        """Test Max t Rq shift power break and no history."""
        A = np.eye(2)
        val, vec = te.max_tRq_shift_power(A, alpha=0.0, maxiter=3, tol=1.0, x0=np.array([1.0, 0.0]), history=False)
        self.assertAlmostEqual(val, 1.0)
        self.assertEqual(vec.shape, (2,))

    def test_max_tRq_geap_break_no_history(self):
        """Test Max t Rq geap break no history."""
        A = np.eye(2)
        val, vec = te.max_tRq_geap(A, maxiter=3, tol=1.0, x0=np.array([1.0, 0.0]), history=False)
        self.assertAlmostEqual(val, 1.0)
        self.assertEqual(vec.shape, (2,))

    def test_tRq_dyn_sys_brute_basic(self):
        """Test T Rq dyn sys brute basic."""
        out = te.tRq_dyn_sys_brute(0.0, np.array([1.0, 0.0]), np.eye(2), d=2, m=2)
        np.testing.assert_array_almost_equal(out, np.zeros(2))

    def test_max_tRq_dynsys_warning_path(self):
        """Test Max t Rq dynsys warning path."""
        A = np.array([[2.0, 0.0], [0.0, 1.0]])

        def fake_odeint(func, y0, t, h0=None, full_output=False, **kwargs):
            """Fake odeint."""
            return np.stack([y0, y0 * 0.5]), {"hu": [h0 or 0.5]}

        with patch("tensiometer.utilities.tensor_eigenvalues.scipy.integrate.odeint", side_effect=fake_odeint):
            val, vec = te.max_tRq_dynsys(A, maxiter=1, tol=0.0, x0=None, h0=0.5, history=False)

        self.assertEqual(vec.shape, (2,))
        self.assertTrue(np.isfinite(val))

    def test_max_tRq_dynsys_break_by_h0_and_history(self):
        """Test Max t Rq dynsys break by h0 and history."""
        A = np.eye(2)

        def fake_odeint(func, y0, t, h0=None, full_output=False, **kwargs):
            """Fake odeint."""
            return np.stack([y0, y0]), {"hu": [h0 or 1.0]}

        with patch("tensiometer.utilities.tensor_eigenvalues.scipy.integrate.odeint", side_effect=fake_odeint):
            val, vec, hist = te.max_tRq_dynsys(
                A, maxiter=1, tol=1.0, x0=np.array([1.0, 0.0]), h0=1.0, history=True
            )

        self.assertEqual(vec.shape, (2,))
        self.assertEqual(hist.shape[0], 1)
        self.assertAlmostEqual(val, 1.0)

    def test_max_tRq_dynsys_h0_limit_break(self):
        """Test Max t Rq dynsys h0 limit break."""
        A = np.eye(2)

        def fake_odeint(func, y0, t, h0=None, full_output=False, **kwargs):
            """Fake odeint."""
            return np.stack([y0, y0]), {"hu": [h0]}

        with patch("tensiometer.utilities.tensor_eigenvalues.scipy.integrate.odeint", side_effect=fake_odeint):
            val, vec = te.max_tRq_dynsys(
                A, maxiter=2, tol=0.0, x0=np.array([0.5, 0.5]), h0=2e6, history=False
            )

        self.assertEqual(vec.shape, (2,))
        self.assertTrue(np.isfinite(val))

    def test_GtRq_Jac_brute(self):
        """Test Gt Rq Jac brute."""
        res = te.GtRq_Jac_brute(
            x=np.array([1.0, 0.0]),
            m=2,
            Axm=2.0,
            Bxm=1.0,
            Axmm1=np.array([2.0, 0.0]),
            Bxmm1=np.array([1.0, 0.0]),
        )
        np.testing.assert_array_almost_equal(res, np.array([4.0, 0.0]))

    def test_max_GtRq_brute_trustregions_branch(self):
        """Test Max Gt Rq brute trustregions branch."""
        class DummySolver:
            """Dummy Solver test suite."""
            def __init__(self, *args, **kwargs):
                """Init."""
                pass

            def solve(self, problem):
                """Solve."""
                return np.array([1.0, 0.0])

        with patch.object(te.pymanopt.solvers, "TrustRegions", DummySolver):
            val, vec = te.max_GtRq_brute(np.eye(2), np.eye(2), optimizer="TrustRegions")

        self.assertTrue(np.isfinite(val))
        self.assertEqual(vec.shape, (2,))

    def test_max_GtRq_brute_particleswarm_branch(self):
        """Test Max Gt Rq brute particleswarm branch."""
        class DummySolver:
            """Dummy Solver test suite."""
            def __init__(self, *args, **kwargs):
                """Init."""
                pass

            def solve(self, problem):
                """Solve."""
                return np.array([1.0, 0.0])

        with patch.object(te.pymanopt.solvers, "ParticleSwarm", DummySolver):
            val, vec = te.max_GtRq_brute(np.eye(2), np.eye(2), optimizer="ParticleSwarm")

        self.assertTrue(np.isfinite(val))
        self.assertEqual(vec.shape, (2,))

    def test_max_GtRq_geap_power_history_and_warning(self):
        """Test Max Gt Rq geap power history and warning."""
        A = np.array([[2.0, 0.0], [0.0, 1.0]])
        # maxiter None triggers sys.maxsize fallback and history return
        lam, vec, hist = te.max_GtRq_geap_power(A, np.eye(2), maxiter=None, tol=1.0, history=True)
        self.assertTrue(np.isfinite(lam))
        self.assertEqual(vec.shape, (2,))
        self.assertIsInstance(hist, np.ndarray)

        # small maxiter enforces warning path
        lam2, vec2 = te.max_GtRq_geap_power(
            A, np.eye(2), maxiter=1, tol=0.0, x0=np.array([1.0, 1.0]), history=False
        )
        self.assertTrue(np.isfinite(lam2))
        self.assertEqual(vec2.shape, (2,))

    def test_max_GtRq_geap_power_warning_banner(self):
        """Test Max Gt Rq geap power warning banner."""
        A = np.array([[2.0, 0.0], [0.0, 1.0]])
        lam, vec = te.max_GtRq_geap_power(
            A, np.eye(2), maxiter=10002, tol=-1.0, x0=np.array([0.6, 0.8]), history=False
        )
        self.assertTrue(np.isfinite(lam))
        self.assertEqual(vec.shape, (2,))

    def test_invalid_optimizer_paths_raise(self):
        """Test Invalid optimizer paths raise."""
        with self.assertRaises(NameError):
            te.max_tRq_brute(np.eye(1), optimizer="Unknown")

        with self.assertRaises(NameError):
            te.max_GtRq_brute(np.eye(1), np.eye(1), optimizer="Unknown")


#########################################################################################################
# Script entry point

if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)
